#!/usr/bin/env python3
"""Functions to set up a finite-volume solution matrix for a landlab grid."""

import numpy as np
from scipy.sparse import csc_matrix

from ._matrix import (
    fill_right_hand_side,
    get_matrix_diagonal_elements,
    get_matrix_diagonal_elements_with_coef,
)


def get_core_node_at_node(grid):
    """Get node ids as numbered by core nodes.

    Get the core node ID for each node of a grid. If a node is not a core
    node, then use -1.

    Parameters
    ----------
    grid : ModelGrid
        A ModelGrid.

    Returns
    -------
    ndarray of int
        Ids of each of the grid's core nodes.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.utils import get_core_node_at_node

    >>> grid = RasterModelGrid((4, 5))
    >>> get_core_node_at_node(grid)
    array([-1, -1, -1, -1, -1,
           -1,  0,  1,  2, -1,
           -1,  3,  4,  5, -1,
           -1, -1, -1, -1, -1])

    >>> grid.status_at_node[13] = grid.BC_NODE_IS_FIXED_VALUE
    >>> grid.status_at_node[2] = grid.BC_NODE_IS_CLOSED
    >>> get_core_node_at_node(grid)
    array([-1, -1, -1, -1, -1,
           -1,  0,  1,  2, -1,
           -1,  3,  4, -1, -1,
           -1, -1, -1, -1, -1])
    """
    core_node_at_node = -np.ones(grid.number_of_nodes, dtype=int)
    core_node_at_node[grid.core_nodes] = np.arange(grid.number_of_core_nodes, dtype=int)
    return core_node_at_node


def get_matrix_entries(grid, coef_at_link=None):
    """Get entries of a sparse matrix.

    Parameters
    ----------
    grid : RasterModelGrid, HexModelGrid
        A landlab grid.
    coef_at_link : ndarray
        Coefficients at links used to construct the matrix.

    Returns
    -------
    tuple of (data, (row_ind, col_inds))
        Values of matrix elements along with their corresponding row and column
        index.
    """
    core2core = grid.link_with_node_status(
        status_at_tail=grid.BC_NODE_IS_CORE, status_at_head=grid.BC_NODE_IS_CORE
    )
    fv2core = grid.link_with_node_status(
        status_at_tail=grid.BC_NODE_IS_FIXED_VALUE, status_at_head=grid.BC_NODE_IS_CORE
    )
    core2fv = grid.link_with_node_status(
        status_at_tail=grid.BC_NODE_IS_CORE, status_at_head=grid.BC_NODE_IS_FIXED_VALUE
    )

    core_node_at_node = get_core_node_at_node(grid)

    nodes_at_c2fv_link = grid.nodes_at_link[core2fv]
    nodes_at_fv2c_link = grid.nodes_at_link[fv2core]

    core_nodes_at_c2c_link = core_node_at_node[grid.nodes_at_link[core2core]]
    core_nodes_at_c2fv_link = core_node_at_node[nodes_at_c2fv_link]
    core_nodes_at_fv2c_link = core_node_at_node[nodes_at_fv2c_link]

    n_core_nodes = grid.number_of_core_nodes

    values = np.zeros(n_core_nodes + 2 * len(core2core), dtype=float)
    row_inds = np.empty(n_core_nodes + 2 * len(core2core), dtype=int)
    col_inds = np.empty(n_core_nodes + 2 * len(core2core), dtype=int)

    diagonal_values = values[:n_core_nodes]
    diagonal_rows = row_inds[:n_core_nodes]
    diagonal_cols = col_inds[:n_core_nodes]

    upper_values = values[n_core_nodes : n_core_nodes + len(core2core)]
    upper_rows = row_inds[n_core_nodes : n_core_nodes + len(core2core)]
    upper_cols = col_inds[n_core_nodes : n_core_nodes + len(core2core)]

    lower_values = values[n_core_nodes + len(core2core) :]
    lower_rows = row_inds[n_core_nodes + len(core2core) :]
    lower_cols = col_inds[n_core_nodes + len(core2core) :]

    if coef_at_link is None:
        get_matrix_diagonal_elements(
            core_nodes_at_c2c_link,
            core_nodes_at_c2fv_link,
            core_nodes_at_fv2c_link,
            diagonal_values,
        )
        upper_values.fill(1.0)
    else:
        get_matrix_diagonal_elements_with_coef(
            core_nodes_at_c2c_link,
            core_nodes_at_c2fv_link,
            core_nodes_at_fv2c_link,
            coef_at_link[core2core],
            coef_at_link[core2fv],
            coef_at_link[fv2core],
            diagonal_values,
        )
        upper_values[:] = coef_at_link[core2core]

    diagonal_rows[:] = np.arange(n_core_nodes)
    diagonal_cols[:] = diagonal_rows

    upper_rows[:] = core_nodes_at_c2c_link[:, 0]
    upper_cols[:] = core_nodes_at_c2c_link[:, 1]

    lower_values[:] = upper_values
    lower_rows[:] = upper_cols
    lower_cols[:] = upper_rows

    return values, (row_inds, col_inds)


def get_core_node_matrix(grid, value_at_node, coef_at_link=None):
    """A matrix for core nodes and a right-hand side vector.

    Construct and return a matrix for the core nodes, plus a right-hand side vector
    containing values based on the input array `value_at_node`. Optionally,
    `coef_at_link` provides coefficients for each link (default is 1.0).

    Parameters
    ----------
    grid : RasterModelGrid, HexModelGrid
        A landlab grid.
    value_at_node : ndarray
        Values defined at nodes used to construct the right-hand side vector.
    coef_at_link : ndarray, optional
        Coefficents at links used to construct the matrix. If not provided,
        use 1.0.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.utils import get_core_node_matrix

    >>> grid = RasterModelGrid((4, 5))
    >>> grid.status_at_node[13] = grid.BC_NODE_IS_FIXED_VALUE
    >>> grid.status_at_node[2] = grid.BC_NODE_IS_CLOSED

    >>> vals = np.arange(grid.number_of_nodes, dtype=np.double)  # made-up state variable array

    >>> mat, rhs = get_core_node_matrix(grid, vals)
    >>> mat.toarray()
    array([[-4.,  1.,  0.,  1.,  0.],
           [ 1., -3.,  1.,  0.,  1.],
           [ 0.,  1., -4.,  0.,  0.],
           [ 1.,  0.,  0., -4.,  1.],
           [ 0.,  1.,  0.,  1., -4.]])
    >>> rhs
    array([[ -6.],
           [  0.],
           [-25.],
           [-26.],
           [-30.]])

    >>> coefs = np.arange(grid.number_of_links, dtype=np.double)  # coefficient array
    >>> mat, rhs = get_core_node_matrix(grid, vals, coef_at_link=coefs)
    >>> mat.toarray()
    array([[-38.,  10.,   0.,  14.,   0.],
           [ 10., -36.,  11.,   0.,  15.],
           [  0.,  11., -46.,   0.,   0.],
           [ 14.,   0.,   0., -74.,  19.],
           [  0.,  15.,   0.,  19., -78.]])
    >>> rhs
    array([[ -6.],
           [  0.],
           [-25.],
           [-26.],
           [-30.]])
    """
    value_at_node = np.broadcast_to(value_at_node, grid.number_of_nodes)
    if coef_at_link is not None:
        coef_at_link = np.broadcast_to(coef_at_link, grid.number_of_links)

    (values, (row_inds, col_inds)) = get_matrix_entries(grid, coef_at_link=coef_at_link)

    mat = csc_matrix(
        (values, (row_inds, col_inds)),
        shape=(grid.number_of_core_nodes, grid.number_of_core_nodes),
    )

    fv2core = grid.link_with_node_status(
        status_at_tail=grid.BC_NODE_IS_FIXED_VALUE, status_at_head=grid.BC_NODE_IS_CORE
    )
    core2fv = grid.link_with_node_status(
        status_at_tail=grid.BC_NODE_IS_CORE, status_at_head=grid.BC_NODE_IS_FIXED_VALUE
    )

    core_node_at_node = get_core_node_at_node(grid)
    nodes_at_c2fv_link = grid.nodes_at_link[core2fv]
    nodes_at_fv2c_link = grid.nodes_at_link[fv2core]

    rhs = np.zeros(grid.number_of_core_nodes, dtype=float)
    fill_right_hand_side(
        nodes_at_c2fv_link, nodes_at_fv2c_link, core_node_at_node, value_at_node, rhs
    )

    return mat, rhs.reshape((-1, 1))



#! /usr/bin/env python
"""
Unit tests for landlab.components.gravel_river_transporter.gravel_river_transporter
"""
from numpy.testing import assert_allclose, assert_equal, assert_raises

from landlab import HexModelGrid, RadialModelGrid, RasterModelGrid
from landlab.components import FlowAccumulator, GravelRiverTransporter


def test_analytical_solution_one_cell_basic_solver():
    # parameters
    dx = 1000.0  # node spacing, m
    uplift_rate = 0.0001  # rate of relative uplift, m/y
    bankfull_runoff_rate = 10.0  # bankfull runoff rate, m/y
    abrasion_coef = 1.0 / 2000.0  # abrasion coefficient, 1/m
    trans_coef = 0.041  # transport coefficient, -
    intermittency_fac = 0.01  # intermittency factor, -
    porosity = 0.35
    dt = 10000.0  # time step duration, y

    grid = RasterModelGrid((3, 3), xy_spacing=dx)
    grid.status_at_node[grid.perimeter_nodes] = grid.BC_NODE_IS_CLOSED
    grid.status_at_node[5] = grid.BC_NODE_IS_FIXED_VALUE

    elev = grid.add_zeros("topographic__elevation", at="node")

    fa = FlowAccumulator(grid, runoff_rate=bankfull_runoff_rate)
    fa.run_one_step()
    transporter = GravelRiverTransporter(
        grid,
        intermittency_factor=intermittency_fac,
        transport_coefficient=trans_coef,
        abrasion_coefficient=abrasion_coef,
        sediment_porosity=porosity,
        solver="explicit",
    )

    for _ in range(200):
        fa.run_one_step()
        elev[grid.core_nodes] += uplift_rate * dt
        transporter.run_one_step(dt)

    # Analytical solutions for sediment flux, slope, and abrasion rate
    Qs_pred = (uplift_rate * dx * dx * (1.0 - porosity)) / (abrasion_coef * dx / 2 + 1)
    S_pred = (
        Qs_pred / (trans_coef * intermittency_fac * bankfull_runoff_rate * dx * dx)
    ) ** (6.0 / 7.0)
    Ea_pred = abrasion_coef * (Qs_pred / 2.0) / dx  # factor of 2 is cell midpoint

    assert_allclose(transporter._sediment_outflux[4], Qs_pred, rtol=1.0e-4)
    assert_allclose(transporter._slope[4], S_pred, rtol=1.0e-4)
    assert_allclose(transporter._abrasion[4], Ea_pred, rtol=1.0e-4)


def test_analytical_solution_one_cell_matrix_solver():
    # parameters
    dx = 1000.0  # node spacing, m
    uplift_rate = 0.0001  # rate of relative uplift, m/y
    bankfull_runoff_rate = 10.0  # bankfull runoff rate, m/y
    abrasion_coef = 1.0 / 2000.0  # abrasion coefficient, 1/m
    trans_coef = 0.041  # transport coefficient, -
    intermittency_fac = 0.01  # intermittency factor, -
    porosity = 0.35
    dt = 10000.0  # time step duration, y

    grid = RasterModelGrid((3, 3), xy_spacing=dx)
    grid.status_at_node[grid.perimeter_nodes] = grid.BC_NODE_IS_CLOSED
    grid.status_at_node[5] = grid.BC_NODE_IS_FIXED_VALUE

    elev = grid.add_zeros("topographic__elevation", at="node")

    fa = FlowAccumulator(grid, runoff_rate=bankfull_runoff_rate)
    fa.run_one_step()
    transporter = GravelRiverTransporter(
        grid,
        intermittency_factor=intermittency_fac,
        transport_coefficient=trans_coef,
        abrasion_coefficient=abrasion_coef,
        sediment_porosity=porosity,
        solver="matrix",
    )

    for _ in range(200):
        fa.run_one_step()
        elev[grid.core_nodes] += uplift_rate * dt
        transporter.run_one_step(dt)

    # Analytical solution for slope (matrix solver does not compute sediment
    # flux or abrasion rate)
    Qs_pred = (uplift_rate * dx * dx * (1.0 - porosity)) / (abrasion_coef * dx / 2 + 1)
    S_pred = (
        Qs_pred / (trans_coef * intermittency_fac * bankfull_runoff_rate * dx * dx)
    ) ** (6.0 / 7.0)
    assert_allclose(transporter._slope[4], S_pred, rtol=1.0e-4)


def test_analytical_solution_four_cells_basic_solver():
    # parameters
    dx = 1000.0  # node spacing, m
    initial_slope = 0.0001  # starting topographic slope, m/m
    bankfull_runoff_rate = 10.0  # r, m/y
    trans_coef = 0.041  # this is kQ
    intermittency_fac = 0.01  # this is I
    abrasion_length_scale = 2000.0  # this is the inverse of beta, in m
    dt = 10000.0  # time step, years
    uplift_rate = 0.0001  # U, m/y
    porosity = 0.35  # lambda_p
    nsteps = 400  # number of time steps

    # Derived parameters
    abrasion_coef = 1.0 / abrasion_length_scale  # this is beta, 1/m

    grid = RasterModelGrid((3, 6), xy_spacing=dx)
    grid.status_at_node[grid.perimeter_nodes] = grid.BC_NODE_IS_CLOSED
    grid.status_at_node[11] = grid.BC_NODE_IS_FIXED_VALUE

    elev = grid.add_zeros("topographic__elevation", at="node")
    elev[:] = initial_slope * (grid.x_of_node.max() - grid.x_of_node)

    fa = FlowAccumulator(grid, runoff_rate=bankfull_runoff_rate)
    fa.run_one_step()
    transporter = GravelRiverTransporter(
        grid, abrasion_coefficient=abrasion_coef, solver="explicit"
    )

    for _ in range(nsteps):
        fa.run_one_step()
        elev[grid.core_nodes] += uplift_rate * dt
        transporter.run_one_step(dt)

    # Compare with analytical solutions in upper cell
    Qs_pred = (uplift_rate * dx * dx * (1.0 - porosity)) / (abrasion_coef * dx / 2 + 1)
    assert_allclose(transporter._sediment_outflux[7], Qs_pred, rtol=1.0e-3)
    S_pred = (
        Qs_pred / (trans_coef * intermittency_fac * bankfull_runoff_rate * dx * dx)
    ) ** (6.0 / 7.0)
    assert_allclose(transporter._slope[7], S_pred, rtol=1.0e-3)
    Ea_pred = abrasion_coef * (Qs_pred / 2) / dx
    assert_allclose(transporter._abrasion[7], Ea_pred, rtol=1.0e-3)


def test_analytical_solution_four_cells_matrix_solver():
    # parameters
    dx = 1000.0  # node spacing, m
    initial_slope = 0.0001  # starting topographic slope, m/m
    bankfull_runoff_rate = 10.0  # r, m/y
    trans_coef = 0.041  # this is kQ
    intermittency_fac = 0.01  # this is I
    abrasion_length_scale = 2000.0  # this is the inverse of beta, in m
    dt = 10000.0  # time step, years
    uplift_rate = 0.0001  # U, m/y
    porosity = 0.35  # lambda_p
    nsteps = 400  # number of time steps

    # Derived parameters
    abrasion_coef = 1.0 / abrasion_length_scale  # this is beta, 1/m

    grid = RasterModelGrid((3, 6), xy_spacing=dx)
    grid.status_at_node[grid.perimeter_nodes] = grid.BC_NODE_IS_CLOSED
    grid.status_at_node[11] = grid.BC_NODE_IS_FIXED_VALUE

    elev = grid.add_zeros("topographic__elevation", at="node")
    elev[:] = initial_slope * (grid.x_of_node.max() - grid.x_of_node)

    fa = FlowAccumulator(grid, runoff_rate=bankfull_runoff_rate)
    fa.run_one_step()
    transporter = GravelRiverTransporter(
        grid, abrasion_coefficient=abrasion_coef, solver="matrix"
    )

    for _ in range(nsteps):
        fa.run_one_step()
        elev[grid.core_nodes] += uplift_rate * dt
        transporter.run_one_step(dt)

    # Compare with analytical solution for slope in upper cell
    Qs_pred = (uplift_rate * dx * dx * (1.0 - porosity)) / (abrasion_coef * dx / 2 + 1)
    S_pred = (
        Qs_pred / (trans_coef * intermittency_fac * bankfull_runoff_rate * dx * dx)
    ) ** (6.0 / 7.0)
    assert_allclose(transporter._slope[7], S_pred, rtol=1.0e-3)


def test_exception_handling():
    grid = RasterModelGrid((3, 3))
    grid.add_zeros("topographic__elevation", at="node")
    FlowAccumulator(grid)
    assert_raises(
        ValueError, GravelRiverTransporter, grid, solver="not one we actually have"
    )


def test_link_length_handling():
    # Hex grid with fixed and uniform flow-link length
    grid = HexModelGrid((5, 3), spacing=2.0)
    grid.add_zeros("topographic__elevation", at="node")
    fa = FlowAccumulator(grid)
    transporter = GravelRiverTransporter(grid)
    assert_allclose(transporter._flow_link_length_over_cell_area, 0.57735, rtol=1.0e-6)

    # Raster grid with possibility of flow on diagonals
    grid = RasterModelGrid((3, 4), xy_spacing=2.0)
    grid.status_at_node[grid.perimeter_nodes] = grid.BC_NODE_IS_CLOSED
    grid.status_at_node[0] = grid.BC_NODE_IS_FIXED_VALUE
    elev = grid.add_zeros("topographic__elevation", at="node")
    elev[5] = 1.0
    elev[6] = 2.0
    fa = FlowAccumulator(grid, flow_director="FlowDirectorD8")
    fa.run_one_step()
    assert_equal(grid.at_node["flow__link_to_receiver_node"][5:7], [17, 8])
    transporter = GravelRiverTransporter(grid, abrasion_coefficient=1.0)
    transporter.run_one_step(1.0)
    assert_equal(transporter._flow_length_is_variable, True)
    assert_equal(transporter._grid_has_diagonals, True)
    assert_allclose(
        transporter._flow_link_length_over_cell_area, [0.707107, 0.5], rtol=1.0e-6
    )

    # Radial grid with variable link length but no diagonals
    grid = RadialModelGrid(n_rings=1, nodes_in_first_ring=5, spacing=2.0)
    elev = grid.add_zeros("topographic__elevation", at="node")
    elev[2] = 1.0
    elev[0] = -1.0
    fa = FlowAccumulator(grid)
    fa.run_one_step()
    transporter = GravelRiverTransporter(grid, abrasion_coefficient=1.0)
    transporter.run_one_step(1.0)
    assert_equal(transporter._flow_length_is_variable, True)
    assert_equal(transporter._grid_has_diagonals, False)
    assert_allclose(transporter._flow_link_length_over_cell_area, [0.5505524])



"""Function to calculate node statistics in a moving window."""

import numpy as np

from landlab import FieldError


def calculate_window_statistic(
    grid, field, func, search_radius, calc_on_closed_nodes=True, **kwargs
):
    """Calculate a statistic using a function within a search window.

    .. note::

        This only works on grid **nodes** (not other grid elements e.g. links) for
        any :class:`~.ModelGrid` type.

    This utility outputs an array of length equal to the grid's number of
    nodes. Each element of the output array represents the node location in
    the grid. The value of each element is a function of the nodes within the
    search window surrounding that node location (see the model grid diagram
    below).

    The grid below contains six columns and five rows with cell spacing set
    to 10 distance units. This utility iteratively evaluates all nodes in the
    grid. The diagram shows evaluation of node ID 15 (marked ``x``). If the
    search radius is set to 20, twice the cell spacing, each node marked with
    a ``*`` is within the search window.
    ::

        · · * · · ·
        · * * * · ·
        * * x * * ·
        · * * * · ·
        · · * · · ·

    Increasing the search radius to 25 results in the following search window.
    ::

        · * * * · ·
        * * * * * ·
        * * x * * ·
        * * * * * ·
        · * * * · ·

    Decreasing the search radius to 15 results in the following search window.
    ::

        · · · · · ·
        · * * * · ·
        · * x * · ·
        · * * * · ·
        · · · · · ·

    The input field can be any field assigned to grid nodes (e.g.
    "topographic__elevation") and the input function can be any function that
    acts on the input field (e.g. "np.min" to find the minimum). The input
    function may be user defined and may contain any number of inputs, which
    are input as ``kwargs``.

    For example, if the input field is "topographic__elevation" and the input
    function is ``np.ptp`` (peak-to-peak, meaning max minus min value), then the
    output at node 15 will be the maximum elevation within the search window
    minus the minimum elevation within the search window (also known as relief).
    The ``np.percentile`` function, however, requires not only the input field,
    but also an input value to define the "q-th percentile" to be calculated.
    This second input would be added as a ``kwarg`` (e.g. ``q=90``) at the end of
    the inputs for :func:`~calculate_window_statistic`. Both of these scenarios are
    shown in the examples below.

    Parameters
    ----------
    grid : ModelGrid
        A Landlab ModelGrid.
    field : string
        An existing grid field on which to calculate the statistic of interest.
        Must exist in grid.
    func : function
        The function that calculates the window statistic of *field*.
        The first parameter of the function must be the values at nodes within
        the window, which are used used to calculate the statistic for the
        node under evaluation. Additional parameters of the function can be
        passed with ``kwargs``.
    search_radius : float
        Radius of window within which the statistic is calculated.
    calc_on_closed_nodes : boolean, optional
        Toggle calculation over all nodes including closed nodes (``True``) or all
        nodes except closed nodes (``False``).
    kwargs : optional
        Keyword arguments passed to *func* that are additional to the array of
        node values within the search window.

    Returns
    -------
    output : ndarray
        Output array containing the calculated values of the statistic. Same
        length as input field.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> from landlab.utils import window_statistic

    >>> grid = RasterModelGrid((5, 6), xy_spacing=10.0)
    >>> grid.set_closed_boundaries_at_grid_edges(False, True, False, True)
    >>> z = grid.add_zeros("topographic__elevation", at="node")
    >>> z += np.arange(len(z))

    Calculate relief using ``np.ptp`` function.

    >>> relief = calculate_window_statistic(
    ...     grid, "topographic__elevation", np.ptp, search_radius=15
    ... )
    >>> grid.at_node["topographic__elevation"]
    array([ 0.,   1.,   2.,   3.,   4.,   5.,
            6.,   7.,   8.,   9.,  10.,  11.,
           12.,  13.,  14.,  15.,  16.,  17.,
           18.,  19.,  20.,  21.,  22.,  23.,
           24.,  25.,  26.,  27.,  28.,  29.])
    >>> relief
    array([ 7.,   8.,   8.,   8.,   8.,   7.,
           13.,  14.,  14.,  14.,  14.,  13.,
           13.,  14.,  14.,  14.,  14.,  13.,
           13.,  14.,  14.,  14.,  14.,  13.,
            7.,   8.,   8.,   8.,   8.,   7.])

    Calculate relief using ``np.ptp`` function excluding closed nodes.

    >>> relief = calculate_window_statistic(
    ...     grid,
    ...     "topographic__elevation",
    ...     np.ptp,search_radius=15,
    ...     calc_on_closed_nodes=False,
    ... )
    >>> grid.at_node["topographic__elevation"]
    array([ 0.,   1.,   2.,   3.,   4.,   5.,
            6.,   7.,   8.,   9.,  10.,  11.,
           12.,  13.,  14.,  15.,  16.,  17.,
           18.,  19.,  20.,  21.,  22.,  23.,
           24.,  25.,  26.,  27.,  28.,  29.])
    >>> relief
    array([ nan,  nan,  nan,  nan,  nan,  nan,
             7.,   8.,   8.,   8.,   8.,   7.,
            13.,  14.,  14.,  14.,  14.,  13.,
             7.,   8.,   8.,   8.,   8.,   7.,
            nan,  nan,  nan,  nan,  nan,  nan])

    Calculate 90th percentile using ``np.percentile`` function and ``kwargs``.

    >>> perc_90 = calculate_window_statistic(
    ...     grid,
    ...     "topographic__elevation",
    ...     np.percentile,search_radius=15,
    ...     calc_on_closed_nodes=False,
    ...     q=90
    ... )
    >>> grid.at_node["topographic__elevation"]
    array([ 0.,   1.,   2.,   3.,   4.,   5.,
            6.,   7.,   8.,   9.,  10.,  11.,
           12.,  13.,  14.,  15.,  16.,  17.,
           18.,  19.,  20.,  21.,  22.,  23.,
           24.,  25.,  26.,  27.,  28.,  29.])
    >>> perc_90
    array([ nan,  nan,  nan,  nan,  nan,  nan,
           12.7, 13.5, 14.5, 15.5, 16.5, 16.7,
           18.5, 19.2, 20.2, 21.2, 22.2, 22.5,
           18.7, 19.5, 20.5, 21.5, 22.5, 22.7,
            nan,  nan,  nan,  nan,  nan,  nan])

    Calculate relief above 90th percentile elevation using a user-defined
    function and ``kwargs``.

    >>> def max_minus_percentile(elev, q):
    ...     output = np.max(elev) - np.percentile(elev, q)
    ...     return output
    >>> rel_above_90th_perc = calculate_window_statistic(
    ...     grid,
    ...     "topographic__elevation",
    ...     max_minus_percentile,
    ...     search_radius=15,
    ...     calc_on_closed_nodes=False,
    ...     q=90,
    ... )
    >>> grid.at_node["topographic__elevation"]
    array([ 0.,   1.,   2.,   3.,   4.,   5.,
            6.,   7.,   8.,   9.,  10.,  11.,
           12.,  13.,  14.,  15.,  16.,  17.,
           18.,  19.,  20.,  21.,  22.,  23.,
           24.,  25.,  26.,  27.,  28.,  29.])
    >>> rel_above_90th_perc
    array([ nan,  nan,  nan,  nan,  nan,  nan,
            0.3,  0.5,  0.5,  0.5,  0.5,  0.3,
            0.5,  0.8,  0.8,  0.8,  0.8,  0.5,
            0.3,  0.5,  0.5,  0.5,  0.5,  0.3,
            nan,  nan,  nan,  nan,  nan,  nan])
    """
    if field not in grid.at_node:
        raise FieldError(f"A {field} field is required at the nodes of the input grid.")

    # Create output array
    output = np.zeros(grid.number_of_nodes)

    # Create arrays of x and y coords for input to "distance to point' calc
    x_coord = grid.x_of_node
    y_coord = grid.y_of_node

    nodes_in_loop = grid.nodes.flatten()
    nodes_to_include = np.ones(grid.number_of_nodes, dtype=bool)

    if calc_on_closed_nodes is False:
        closed_nodes = grid.status_at_node == grid.BC_NODE_IS_CLOSED
        nodes_in_loop = nodes_in_loop[~closed_nodes]
        nodes_to_include[closed_nodes] = False
        output[closed_nodes] = np.NaN

    # Calculate "dist to point" then local value at nodes within window.
    for node in nodes_in_loop:
        node_dist_to_point = grid.calc_distances_of_nodes_to_point(
            (x_coord[node], y_coord[node])
        )
        nodes_in_window = np.all(
            [node_dist_to_point <= search_radius, nodes_to_include], 0
        )
        values_in_window = grid.at_node[field][nodes_in_window]
        output[node] = func(values_in_window, **kwargs)

    return output



import pytest

from landlab import RasterModelGrid
from landlab.components import SedimentPulserAtLinks, SedimentPulserEachParcel

variable_list = {
    "_grid",
    "_parcels",
    "_D50",
    "_D84_D50",
    "_rho_sediment",
    "_parcel_volume",
    "_abrasion_rate",
}


def test_basic_init_each_parcel(example_nmg, example_parcels):
    """test all class variables are present after SedimentPulserEachParcel initializes"""
    pulser = SedimentPulserEachParcel(example_nmg, parcels=example_parcels)
    assert variable_list.issubset(vars(pulser))


def test_basic_init_at_link(example_nmg, example_parcels):
    """test all class variables are present after SedimentPulserAtLinks initializes"""
    pulser = SedimentPulserAtLinks(example_nmg, parcels=example_parcels)
    assert variable_list.issubset(vars(pulser))


def test_grid_is_nmg():
    """test ValueError exception is raised when nmg is a mg"""
    nmg = RasterModelGrid((5, 5))
    with pytest.raises(ValueError):
        _ = SedimentPulserAtLinks(nmg)
    with pytest.raises(ValueError):
        _ = SedimentPulserEachParcel(nmg)



#!/usr/bin/env python
"""ZoneController of SpeciesEvolver."""
import numpy as np
from scipy.ndimage.measurements import label

from .record import Record
from .zone import Zone, _update_zones
from .zone_taxon import ZoneTaxon


class ZoneController:
    """Controls zones and populates them with taxa.

    This object manages 'zones' that are used to evaluate the spatial aspect of
    taxa. A zone represents a portion of a model grid. It is made up of
    spatially continuous grid nodes.

    This controller creates zones using the initialization parameter,
    ``zone_function``. This function identifies all of the grid nodes where
    zones are to be created. A zone is created for each cluster of spatially
    continuous nodes. Zones are updated also using this function when the
    ``run_one_step`` method of this controller is called.

    The structure of an example model grid is diagrammed below to demonstrate
    how zones are created. The grid contains six columns and six rows. In this
    example, the function returns True where node values are greater than 0.
    Nodes marked with an ``*`` are nodes that will belong to a zone because the
    mask is True at these nodes. All other nodes are marked with a ``·``. A
    zone is created for each cluster of continuous nodes where the mask is True.
    ::

        values         mask returned
        evaluated      by zone function
        0 0 0 0 0 0    · · · · · ·
        0 0 0 5 4 2    · · · * * *
        0 6 0 4 6 0    · * · * * ·
        0 2 0 0 0 0    · * · · · ·
        0 0 4 0 4 0    · · * · * ·
        0 0 0 0 0 0    · · · · · ·

    The above example is continued in the following four diagrams that
    demonstrate how individual zones are identified. Each zone is marked with a
    ``x``, ``o``, ``+``, or ``@``. Clusters can be identified using ``D8``
    where diagonal neighbors are included or ``D4`` where diagonal neighbors
    are excluded. A minimum zone area can be enforced with the ``minimum_area``
    initialization parameter.
    ::

        D8             D4             D8             D4
        min area = 0   min area = 0   min area = 2   min area = 2
        · · · · · ·    · · · · · ·    · · · · · ·    · · · · · ·
        · · · + + +    · · · + + +    · · · + + +    · · · + + +
        · x · + + ·    · x · + + ·    · x · + + ·    · x · + + ·
        · x · · · ·    · x · · · ·    · x · · · ·    · x · · · ·
        · · x · o ·    · · @ · o ·    · · x · · ·    · · · · · ·
        · · · · · ·    · · · · · ·    · · · · · ·    · · · · · ·

    The grid perimeter affects zone creation. Zones can include perimeter nodes
    (the nodes along grid edges), although because perimeter nodes are not
    associated with cells that have area, perimeter nodes do not contribute to
    the area summation of clusters. The area summation only takes into account
    the cells associated with core nodes.

    Creation of zones along boundaries is illustrated below. A zone extent mask
    different from the above example was produced by the hypothetical zone
    function in this example. Again ``*`` indicates where a zone can exist.
    Distinct zones include the symbols, ``$`` and ``#`` in addition to the
    symbols defined above. Individual zone masks and the count of zones are
    affected by the use of ``D8`` or ``D4`` along with the minimum area
    parameter, especially when zone clusters are along the grid parameter.
    ::

        zone function  D8             D4             D8             D4
        returned mask  min area = 0   min area = 0   min_area = 2   min_area = 2
        * · · * * ·    + · · x x ·    + · · x x ·    + · · · · ·    · · · · · ·
        · * · · * ·    · + · · x ·    · # · · x ·    · + · · · ·    · # · · · ·
        · * · · · ·    · + · · · ·    · # · · · ·    · + · · · ·    · # · · · ·
        * · · · * *    + · · · o o    $ · · · o o    + · · · o o    · · · · o o
        · · · · * ·    · · · · o ·    · · · · o ·    · · · · o ·    · · · · o ·
        · * * · · ·    · @ @ · · ·    · @ @ · · ·    · · · · · ·    · · · · · ·

    By default, ``ZoneTaxon`` are used with this controller, and the
    following paragraphs make that assumption. See the documentation of the
    populate methods to learn how to use other types. Speciation of
    ``ZoneTaxon`` objects occurs when the taxon exists in more than one zone
    once the allopatric wait time has been exceeded in that zone. See
    ``ZoneTaxon`` documentation for more about allopatric wait time.

    A different example grid demonstrates here the temporal connectivity of
    zones. The grid represents the time, ``T0`` with the nodes of a zone
    marked with ``x``. The following examples will use D8 neighborhoods and a
    minimum zone area of 0.
    ::

        T0
        · · · · · ·
        · · · · · ·
        · x x x x ·
        · x x x x ·
        · · · · · ·
        · · · · · ·

    Below are variations of the grid at a later time, ``T1`` in four variations
    where each contains one zone. In ``T1a``, ``T1b``, and ``T1c`` the zone
    stayed the same, moved, and changed size, respectively. Taxa migrate with
    the zone when at least one zone node overlaps between the two time steps.
    However, in ``T1d``, no nodes overlaps, therefore taxa do not disperse from
    the zone in T0 to the zone in T1d.
    ::

        T1a            T1b            T1c            T1d
        · · · · · ·    · · · · · ·    · · · · · ·    · · · · · ·
        · + + + + ·    · · · · · ·    · · + + · ·    · · · · · ·
        · + + + + ·    · + + + + ·    · + + + + ·    · · · · · ·
        · · · · · ·    · + + + + ·    · · · · + ·    · + + + + ·
        · · · · · ·    · · · · · ·    · · · · · ·    · + + + + ·
        · · · · · ·    · · · · · ·    · · · · · ·    · · · · · ·

    Another ``T1`` variation, now demonstrating two zones, ``+`` and ``x``.
    Multiple zones overlapping a zone in the prior time step can be interpreted
    as a zone that fragmented, which may affect resident taxa. The number of
    zone fragmentations can be viewed in the ``record_data_frame`` attribute.
    In the T1e example, the fragmentation count for time 1 would be 2 because
    2 zones that fragmented from a prior zone were recognized at this time.
    ::

        T1e
        · · · · · ·
        · · · · + ·
        · x · · + ·
        · x x · · ·
        · x x · · ·
        · · · · · ·

    The controller had to decide which of the two clusters of continuous nodes
    in T1 should be designated as the same zone in T0. This decision must be
    made in general when multiple clusters overlap the same zone in the prior
    time step. The zone in the current time step that overlaps the prior time
    step zone the most becomes the same zone in the earlier time step. In this
    example, the cluster to the right overlapped four nodes and the left
    cluster overlapped only one node, therefore the right cluster became the
    star zone. This is merely for creating new zones objects.

    The grid diagrammed below continues from T1e. The continuous nodes
    overlapped two zones in T1e. When multiple zones overlap, one zone is
    assumed to be the prior zone and the others are considered captured zones.
    The number of zone captures can be viewed in the ``record_data_frame``
    attribute.
    ::

        T2
        · · · · · ·
        · · · · · ·
        · x x x x ·
        · x x x · ·
        · · · · · ·
        · · · · · ·

    The controller had to again decide which of the two clusters of continuous
    nodes in T1e should be designated as the same zone in T2. This decision
    must be made in general when multiple clusters in the prior time step
    overlap a zone in the current time step. The zone in the prior time step
    that overlaps the current time step zone the most becomes the zone in the
    earlier time step. In this example, the cluster to the left overlapped two
    nodes and the right cluster overlapped only one node, therefore the new
    zone keeps the designation of the left cluster. However, this is merely for
    creating new zone objects.

    ZoneController is currently designed for use with only the grid type,
    ``RasterModelGrid``.

    Examples
    --------
    Import modules used in the following examples.

    >>> from landlab import RasterModelGrid
    >>> from landlab.components.species_evolution import ZoneController

    The first example uses the default parameters of ZoneController.

    Create a model grid and an elevation field for this grid.

    >>> mg = RasterModelGrid((3, 7))
    >>> z = mg.add_zeros('topographic__elevation', at='node')

    Set elevation to 1 for some nodes.

    >>> z[[9, 10, 11, 12]] = 1

    Define a zone function that returns a boolean array where `True` values
    indicate the nodes where zones can be created.

    >>> def zone_func(grid):
    ...     z = grid.at_node['topographic__elevation']
    ...     return z == 1

    Instantiate ZoneController. Only one zone exists because the nodes that
    were set to one are adjacent to each other in the grid.

    >>> zc = ZoneController(mg, zone_func)
    >>> zc.record_data_frame[['time', 'zones']]
       time  zones
    0     0      1

    Populate each zone with a taxon.

    >>> taxon = zc.populate_zones_uniformly(1)
    >>> len(taxon)
    1

    A change in elevation is forced to demonstrate a zone fragmentation, and
    then the zones are updated by advancing the record time by 1000.

    >>> z[10] = 0
    >>> zc.run_one_step(1000)

    Two zones now exist because the zone in time 0 fragmented into two zones.

    >>> zc.record_data_frame[['time', 'zones', 'fragmentations']]
       time  zones  fragmentations
    0     0      1             NaN
    1  1000      2             2.0

    A change in elevation is forced again, this time to demonstrate zone
    capture where multiple zones are overlapped by a zone in the later time
    step. Statistics of the capture can be attained with ``record_data_frame``.

    >>> z[10] = 1
    >>> zc.run_one_step(1000)
    >>> zc.record_data_frame[['time', 'zones', 'captures',
    ...     'area_captured_sum', 'area_captured_max']]
       time  zones  captures  area_captured_sum  area_captured_max
    0     0      1       NaN                NaN                NaN
    1  1000      2       0.0                0.0                0.0
    2  2000      1       1.0                2.0                2.0

    The follow example demonstrates non-default ZoneController parameters.

    >>> mg = RasterModelGrid((3, 7))
    >>> z = mg.add_zeros('topographic__elevation', at='node')

    Similar to the prior example, define a zone function that returns a boolean
    array where `True` values indicate the nodes where zones can be created.

    >>> def zone_func(grid):
    ...     z = grid.at_node['topographic__elevation']
    ...     return z == 1

    Set elevation to 1 for nodes so that two clusters of nodes within the zone
    mask exist.

    >>> z[[9, 10, 12]] = 1

    Instantiate ZoneController with options.

    >>> zc = ZoneController(mg, zone_func, minimum_area=2, initial_time=100)

    Only one zone exist, despite two clusters of nodes meeting the zone
    definition, because the ``minimum_area`` was set to 2. Also, the first
    time in the record was set by the ``initial_time`` parameter.

    >>> zc.record_data_frame[['time', 'zones']]
       time  zones
    0   100      1
    """

    def __init__(
        self,
        grid,
        zone_function,
        minimum_area=0,
        neighborhood_structure="D8",
        initial_time=0,
        **kwargs,
    ):
        """Initialize the controller.

        Parameters
        ----------
        grid : RasterModelGrid
            A Landlab RasterModelGrid.
        zone_function : function
            A function that return a mask of the total zone extent. The first
            input parameter of this function must be `grid`.
        minimum_area : float, optional
            The minimum area of the zones that will be created.
        neighborhood_structure : {'D8', 'D4'}, optional
            The structure describes how zones are identified. The default,
            'D8' evaluates the eight neighboring nodes. The diagonal
            neighboring nodes are excluded when 'D4' is selected.
        initial_time : float, int, optional
            The initial time. The unit of time is unspecified within the
            controller. The default is 0.
        kwargs
            Keyword arguments for ``zone_function``. Do not include ``grid``
            in kwargs because ``grid``, the first parameter of this method, is
            automatically added to ``kwargs``.
        """
        # Set parameters.

        self._grid = grid
        self._zone_func = zone_function
        self._zone_params = kwargs
        self._min_area = minimum_area
        self._record = Record(initial_time)
        if neighborhood_structure in ["D8", "D4"]:
            self._neighborhood_struct = neighborhood_structure
        else:
            raise ValueError("`neighborhood_structure` must be 'D8' or 'D4'")

        # Set record initial values.

        self._record.set_value("zones", np.nan)
        self._record.set_value("fragmentations", np.nan)
        self._record.set_value("captures", np.nan)
        self._record.set_value("area_captured_sum", np.nan)
        self._record.set_value("area_captured_max", np.nan)

        # Include `grid` in the zone params dictionary.

        self._zone_params["grid"] = self._grid

        # Set initial zones.

        initial_zone_extent = self._zone_func(**self._zone_params)
        self._zones = self._get_zones_with_mask(initial_zone_extent)

        self._record.set_value("zones", len(self._zones))

    @property
    def zones(self):
        """The zones of the ZoneController."""
        return self._zones

    @property
    def record_data_frame(self):
        """A DataFrame of ZoneController variables over time.

        Each row is data of a model time step. The step time is recorded in the
        `time` column. The columns, `zones`, `fragmentations`, and `captures`
        are the count of these variables at a given time. `area_captured_sum`
        is the summation of captures over a time. `area_captured_max` is the
        maximum area captured of a single capture during a time.
        """
        return self._record.data_frame

    def populate_zones_uniformly(self, count, taxon_type=ZoneTaxon, **kwargs):
        """Populate each zone with the same type and count of taxa.

        Parameters
        ----------
        count : int
            The count of taxon to populate to each zone.
        taxon_type : type of Taxon
            A Taxon type that takes a Zone as its first parameter.
        kwargs : dictionary
            Keyword arguments of ``taxon_type``.
        """
        taxa = []

        for z in self._zones:
            taxa.extend([taxon_type([z], **kwargs) for _ in range(count)])

        return taxa

    def run_one_step(self, dt):
        """Update the zones for a single timestep.

        This method advances time in the record and determines the connectivity
        of zones between the current and prior time steps.

        Parameters
        ----------
        dt : float
            The model time step duration.
        """
        self._record.advance_time(dt)

        # Resolve the spatiotemporal connectivity of the prior time step zones
        # to the new zones.

        prior_zones = self._zones

        zone_mask = self._zone_func(**self._zone_params)
        new_zones = self._get_zones_with_mask(zone_mask)

        self._zones = _update_zones(self._grid, prior_zones, new_zones, self._record)

        self._record.set_value("zones", len(self._zones))

    def _get_zones_with_mask(self, mask):
        """Get zones using a mask.

        Parameters
        ----------
        mask : ndarray
            A boolean array with the grid number of nodes where `True` values
            are nodes within the extent of all the zones to be created.

        Returns
        -------
        list of Zones
            The discrete zones identified in the mask.
        """
        # Label clusters of `True` values in `mask`.

        if self._neighborhood_struct == "D8":
            s = 3 * [[1, 1, 1]]
        elif self._neighborhood_struct == "D4":
            s = [[0, 1, 0], [1, 1, 1], [0, 1, 0]]

        cluster_arr, cluster_ct = label(mask.reshape(self._grid.shape), structure=s)

        # Create zones for clusters.

        zones = []

        for i in range(1, cluster_ct + 1):
            mask = (cluster_arr == i).flatten()
            cluster_area = self._grid.cell_area_at_node[mask].sum()

            if cluster_area >= self._min_area:
                zones.append(Zone(self, mask))

        return zones



#! /usr/env/python
"""Python implementation of :class:`~.FramedVoronoiGrid`, a grid class used to create and
manage unstructured Voronoi-Delaunay grids for 2D numerical models, with a structured
perimeter layout

Do NOT add new documentation here. Grid documentation is now built in a
semi- automated fashion. To modify the text seen on the web, edit the
files `docs/text_for_[gridfile].py.txt`.

.. codeauthor:: sebastien lenard
"""

import numpy

from ..graph import DualFramedVoronoiGraph
from .base import ModelGrid


class FramedVoronoiGrid(DualFramedVoronoiGraph, ModelGrid):
    """A grid of Voronoi Delaunay cells with a structured perimeter layout.

    This inherited class implements a irregular 2D grid with Voronoi Delaunay cells and
    irregular patches. It is a special type of :class:`~.VoronoiDelaunayGrid` grid in which
    the initial set of points is arranged in a fixed lattice (e.g. like a
    :class:`~.RasterModelGrid`), named here "layout", and the core points are
    then moved a random distance from their initial positions, bounded by a user-supplied
    threshold.

    Examples
    --------
    Create a grid with 3 rows and 2 columns of nodes.

    >>> from landlab import FramedVoronoiGrid
    >>> grid = FramedVoronoiGrid((3, 2), xy_spacing=1.0)
    >>> grid.number_of_nodes
    6

    >>> grid = FramedVoronoiGrid(
    ...     (4, 3), xy_spacing=(10., 10.), xy_min_spacing=(5., 5.), seed=200
    ... )
    >>> grid.status_at_node
    array([1, 1, 1, 1, 0, 1, 1, 0, 1, 1, 1, 1], dtype=uint8)
    >>> grid.x_of_node[3]
    0.0
    >>> grid.x_of_node[5]
    20.0
    >>> grid.y_of_node[0::3]       # doctest: +NORMALIZE_WHITESPACE
    array([  0.   ,   7.499,  17.499,  30.   ])

    >>> grid = FramedVoronoiGrid((3, 5), xy_spacing=(10., 10.), xy_min_spacing=5., seed=None)
    >>> grid.boundary_nodes
    array([ 0,  1,  2,  3,  4,  5,  9, 10, 11, 12, 13, 14])
    """

    # Inheritance diagram:
    #
    # FramedVoronoiGrid
    # |-- ModelGrid
    # `-- DualFramedVoronoiGraph
    #     |-- FramedVoronoiGraph (Layout: HorizontalRectVoronoiGraph)
    #     |   `-- DelaunayGraph
    #     `-- DualGraph (use of static Graph.sort())

    def __init__(
        self,
        shape,
        xy_spacing=(1.0, 1.0),
        xy_of_lower_left=(0.0, 0.0),
        xy_min_spacing=(0.5, 0.5),
        seed=200,
        xy_of_reference=(0.0, 0.0),
        xy_axis_name=("x", "y"),
        xy_axis_units="-",
    ):
        """Create a grid of voronoi cells with a structured perimeter.

        Create an irregular 2D grid with voronoi cells and triangular patches.
        It is a special type of :class:`~.VoronoiDelaunayGrid` in which the initial set
        of points is arranged in a regular lattice determined by the parameters
        *shape*, and *xy_spacing*. The coordinates of
        the core points are then randomly moved while the perimeter points
        remaining fixed, in a way determined by the parameters *xy_min_spacing*, and
        *seed*.

        Parameters
        ----------
        shape : tuple of int
            Number of rows and columns of nodes.
        xy_spacing : float or tuple of float, optional
            Node spacing along x and y coordinates. If ``float``, same spacing at *x* and *y*.
        xy_of_lower_left : tuple, optional
            Minimum *x*-of-node and *y*-of-node values. Depending on the grid,
            there may not be a node at this coordinate.
        xy_min_spacing: float or tuple of float, optional
            Final minimal spacing between nodes. Random moves of the core nodes
            from their initial positions cannot be above this threshold:
            ``(xy_spacing - xy_min_spacing) / 2``
            If ``float``, same minimal spacing for *x* and *y*.
        seed: int, optional
            Seed used to generate the random *x* and *y* moves.
            When set, controls the pseudo-randomness of moves to ensure
            reproducibility.
            When ``None``, the seed is random and the moves of coordinates are
            completely random.
        xy_of_reference : tuple, optional
            Coordinate value in projected space of the reference point,
            *xy_of_lower_left*.
        xy_axis_name: tuple of str, optional
            *x* and *y* axis names.
        xy_axis_units: str, optional
            *x* and *y* axis units.

        Returns
        -------
        FramedVoronoiGrid
            A newly-created grid.

        Examples
        --------
        Create a grid with 3 rows and 2 columns of nodes.

        >>> from landlab import FramedVoronoiGrid
        >>> grid = FramedVoronoiGrid((3, 2), xy_spacing=1.0)
        >>> grid.number_of_nodes
        6
        """
        DualFramedVoronoiGraph.__init__(
            self,
            shape,
            xy_spacing=xy_spacing,
            xy_of_lower_left=xy_of_lower_left,
            sort=True,
            xy_min_spacing=xy_min_spacing,
            seed=seed,
        )
        ModelGrid.__init__(
            self,
            xy_axis_name=xy_axis_name,
            xy_axis_units=xy_axis_units,
            xy_of_reference=xy_of_reference,
        )

        self._node_status = numpy.full(
            self.number_of_nodes, self.BC_NODE_IS_CORE, dtype=numpy.uint8
        )
        self._node_status[self.perimeter_nodes] = self.BC_NODE_IS_FIXED_VALUE

    @classmethod
    def from_dict(cls, kwds):
        args = ()
        return cls(*args, **kwds)



# ! /usr/env/python
"""profiler.py component to create profiles with user-defined endpoints."""
from collections import OrderedDict

import numpy as np
from matplotlib import cm, colors, pyplot as plt

from landlab.components.profiler.base_profiler import _BaseProfiler


class Profiler(_BaseProfiler):
    """Extract and plot profiles set up using points within a grid.

    The profile is constructed from the first to final point in ``endpoints``.
    Endpoints are located at grid nodes. Two successive endpoints bound a
    profile segment. A profile with one segment is a straight line. The
    segments of a profile with multiple segments meet at endpoints. The grid
    nodes along the profile are sampled, including the segment endpoints. The
    extracted quantity of the node is retained. No interpolation is conducted
    even for profile traces that reside between nodes.

    The structure of the profile in a model grid is diagrammed below. The grid
    contains nine columns and nine rows. The profile is constructed from three
    endpoints that bound two segments. Here, ``o`` indicates a segment
    endpoint, ``.`` and ``*`` are sample nodes of the first and second segment,
    respectively. ``X`` are nodes not included in the profile. The first
    segment begins in the lower-left and continues horizontally and almost
    reaches the right boundary. The second segment is joined to the first in
    the lower-right of the grid and it continues diagonally to the upper-left.
    Segments have seven sample points each (nodes at endpoints are also
    sampled). The segments share the second endpoint. Segment and sample
    ordering is dictated by the ordering of endpoints. If the horizontal
    segment is the first segment, the endpoints used to construct this profile
    must be ordered: lower-left, lower-right, and then upper-left.::

        X X X X X X X X X
        X o X X X X X X X
        X X * X X X X X X
        X X X * X X X X X
        X X X X * X X X X
        X X X X X * X X X
        X X X X X X * X X
        X o . . . . . o X
        X X X X X X X X X

    The node IDs and distances along the profile are stored in a data structure
    called ``data_structure``. It is a dictionary with keys indicating the
    segment IDs that are enumerated along the profile.

    By default, a unique color will be assigned to each segment. To change the
    color, a user can change values stored in ``data_structure``. Additionally,
    a ``cmap`` keyword argument can provide some user control over the color at
    the instantiation of the component.

    The data structure of the example above will look as follows:

    .. code-block:: python

        {0: {
            'ids': [10, 11, 12, 13, 14, 15, 16],
            'distances': [0, 1, 2, 3, 4, 5, 6]
            'color': (0.27,  0,  0.33,  1)
            },
         1: {
            'ids': [16, 24, 32, 40, 48, 56, 64],
            'distances': [6, 7.41, 8.83, 10.24, 11.66, 13.07, 14.49]
            'color': (0.13,  0.57,  0.55,  1)
            }
        }


    Examples
    --------
    Create a model grid with the same dimensions as the diagram above.

    >>> from landlab import RasterModelGrid
    >>> from landlab.components import Profiler
    >>> import numpy as np
    >>> mg = RasterModelGrid((10, 10), 10)
    >>> mg.at_node['topographic__elevation'] = mg.node_x * mg.node_y

    Create a profile with three endpoints. This profile is laid out the same as
    the diagram above.

    >>> endpoints = [10, 16, 64]
    >>> profiler = Profiler(mg, endpoints)
    >>> profiler.run_one_step()

    The keys of the data structure are the segment ids.

    >>> profiler.data_structure.keys()
    odict_keys([0, 1])

    The data structure contains data of segment samples. Below is the first
    segment.

    >>> profiler.data_structure[0]['ids']
    array([10, 11, 12, 13, 14, 15, 16])
    >>> profiler.data_structure[0]['distances']
    array([  0.,  10.,  20.,  30.,  40.,  50.,  60.])
    >>> np.round(profiler.data_structure[0]['color'], decimals=2)
    array([ 0.27,  0.  ,  0.33,  1.  ])

    Note that the first node of the second segment is the same as the final
    node of the first segment.

    >>> profiler.data_structure[1]['ids']
    array([16, 26, 35, 45, 54, 64])

    Alternative to nodes, profiles can be instantiated with coordinates.

    >>> profiler = Profiler(mg, [(10, 10), (70, 10), (10, 70)])

    Endpoints can also be set with a combination of coordinates and nodes.

    >>> profiler = Profiler(mg, [(10, 10), 16, (10, 70)])

    References
    ----------
    **Required Software Citation(s) Specific to this Component**

    None Listed

    **Additional References**

    None Listed

    """

    _name = "Profiler"

    _unit_agnostic = True

    def __init__(self, grid, endpoints, cmap="viridis"):
        """Instantiate Profiler.

        Parameters
        ----------
        grid : RasterModelGrid
            A landlab RasterModelGrid.
        endpoints : list of node id integers or coordinate tuples
            The endpoints that bound segments of the profile. Endpoints can be
            node ids and/or tuples of coordinates (x, y, where these
            coordinates are the measurement from the grid lower-left). The list
            can be a mix of node ids and coordinate tuples. The profile begins
            with the first element of `endpoints` and continues in the order of
            this list.
        cmap : str
            A valid matplotlib cmap string. Default is "viridis".
        """
        super().__init__(grid)

        self._cmap = plt.get_cmap(cmap)

        if not isinstance(endpoints, list) or len(endpoints) < 2:
            raise ValueError(
                "`endpoints` must be a list of at least 2 node IDs or a "
                "list of at least two tuples where each tuple contains the "
                "x, y coordinates of endpoints."
            )

        # Check if `endpoints` are within grid bounds while setting
        # `_end_nodes`.

        self._end_nodes = []
        for point in endpoints:
            node, _ = self._get_node_and_coords(point)
            self._end_nodes.append(node)

    @property
    def data_structure(self):
        """OrderedDict defining the profile.

        The node IDs and distances along the profile are stored in
        ``data_structure``. It is a dictionary with keys of the segment ID.
        The value of each key is itself a dictionary of the segment attributes.
        First, 'ids' contains a list of the node IDs of segment samples ordered
        from the start to the end of the segment. It includes the endpoints.
        Second, 'distances' contains a list of along-profile distances that
        mirrors the list in 'ids'. Finally, 'color' is an RGBA tuple indicating
        the color for the segment.
        """
        return self._data_struct

    def _create_profile_structure(self):
        """Create the data structure of the profile.

        The profile is processed by segment. Segments are bound by successive
        endpoints. The cumulative distance along the profile is accumulated by
        iteratively adding segment lengths.
        """
        self._data_struct = OrderedDict()
        grid = self._grid
        endnodes = self._end_nodes
        cum_dist = 0

        for i_endpt in range(len(endnodes) - 1):
            # Get the endpoints and samples of the segment.

            start_node, start_xy = self._get_node_and_coords(endnodes[i_endpt])
            end_node, end_xy = self._get_node_and_coords(endnodes[i_endpt + 1])

            sample_nodes = self._get_sample_nodes(start_node, end_node)

            # Calculate the along-profile distance of samples along the
            # segment.

            n_samples = len(sample_nodes)
            sample_distances = np.empty(n_samples, dtype=float)

            for i_sample, node in enumerate(sample_nodes):
                sample_xy = grid.xy_of_node[node]

                pt = self._project_point_onto_line(sample_xy, start_xy, end_xy)
                d = grid.calc_distances_of_nodes_to_point(pt, node_subset=start_node)
                sample_distances[i_sample] = d

            # Store the segment data.

            self._data_struct[i_endpt] = {
                "ids": np.array(sample_nodes),
                "distances": sample_distances + cum_dist,
            }

            cum_dist += max(sample_distances)

        self._assign_colors()
        self._create_flat_structures()

    def _assign_colors(self, color_mapping=None):
        """Assign a unique color for each segment.

        Parameters
        ----------
        color_mapping : str
            Color map name.
        """
        if color_mapping is None:
            segment_count = len(self._data_struct)
            norm = colors.Normalize(vmin=0, vmax=segment_count)
            mappable = cm.ScalarMappable(norm=norm, cmap=self._cmap)
            color_mapping = {
                segment_id: mappable.to_rgba(idx)
                for idx, segment_id in enumerate(self._data_struct)
            }

        for segment_id in self._data_struct:
            self._data_struct[segment_id]["color"] = color_mapping[segment_id]

    def _create_flat_structures(self):
        """Create expected flattened structures for ids, distances, and colors."""
        self._nodes = []
        self._distance_along_profile = []
        self._colors = []

        for segment_id in self._data_struct:
            self._nodes.append(self._data_struct[segment_id]["ids"])
            self._distance_along_profile.append(
                self._data_struct[segment_id]["distances"]
            )
            self._colors.append(self._data_struct[segment_id]["color"])

    def _get_node_and_coords(self, point):
        """Get the node and coordinates for a point.

        This method handles the option that endpoints can be a node or tuple.
        The grid methods called here verify if the point is within the grid.
        """
        if isinstance(point, (float, int, np.integer)):
            return point, self._grid.xy_of_node[point]
        elif isinstance(point, (tuple, list, np.ndarray)) and len(point) == 2:
            return self._grid.find_nearest_node(point), point
        else:
            raise TypeError(
                "each element of `endpoints` must be a number "
                "representing a node id or a tuple of node x, y "
                "coordinates"
            )

    def _get_sample_nodes(self, start_node, end_node):
        """Get the profile sample nodes using Bresenham's line algorithm.

        Parameters
        ----------
        start_node, end_node : integer
            The node id of a profile endpoint.

        Returns
        -------
        list of integers
            The node ids of the profile samples.

        Notes
        -----
        See: https://en.wikipedia.org/wiki/Bresenham%27s_line_algorithm
        """
        # Get node column and row numbers to act as the coordinates.

        y0, x0 = np.argwhere(self._grid.nodes == start_node)[0]
        y1, x1 = np.argwhere(self._grid.nodes == end_node)[0]

        dx = x1 - x0
        dy = y1 - y0

        trace_is_steep = abs(dy) > abs(dx)

        if trace_is_steep:
            x0, y0 = y0, x0
            x1, y1 = y1, x1

        flipped_nodes = x0 > x1

        if flipped_nodes:
            x0, x1 = x1, x0
            y0, y1 = y1, y0

        dx = x1 - x0
        dy = y1 - y0

        error = int(dx / 2.0)

        if y0 < y1:
            y_step = 1
        else:
            y_step = -1

        # Iterate within the bounding box to identify the profile sample nodes.

        samples = []
        y = y0

        for x in range(x0, x1 + 1):
            if trace_is_steep:
                coord = (x, y)
            else:
                coord = (y, x)

            samples.append(self._grid.grid_coords_to_node_id(*coord))

            error -= abs(dy)
            if error < 0:
                y += y_step
                error += dx

        if flipped_nodes:
            samples.reverse()

        return samples

    def _project_point_onto_line(self, p, ep0, ep1):
        """Get the coordinates along a line nearest to a point.

        Parameters
        ----------
        p : tuple of floats
            The x, y coordinates of the point to project onto the line.
        ep0, ep1 : tuple of floats
            The endpoints of the line. Each endpoint is a tuple of x, y
            coordinates.

        Returns
        -------
        tuple
            The x, y coordinates along a line (bounded by `ep1` and `ep2`) that
            is nearest to `p`.
        """
        dx, dy = ep1[0] - ep0[0], ep1[1] - ep0[1]
        determinant = dx * dx + dy * dy
        coeff = (dy * (p[1] - ep0[1]) + dx * (p[0] - ep0[0])) / determinant

        return ep0[0] + coeff * dx, ep0[1] + coeff * dy

    def plot_profiles_in_map_view(
        self, field="topographic__elevation", endpoints_only=True, **kwds
    ):
        """Plot profile locations in map view.

        This method overrides the method in ``_BaseProfiler`` to set the
        default of ``endpoints_only`` to True.

        Parameters
        ----------
        field : field name or nnode array
            Array of the at-node-field to plot as the 2D map values.
            Default value is the at-node field 'topographic__elevation'.
        endpoints_only : boolean
            Boolean where False indicates every node along the profile is
            plotted, or True (default) indicating only segment endpoints are
            plotted.
        **kwds : dictionary
            Keyword arguments to pass to imshow_grid.
        """
        super().plot_profiles_in_map_view(field, endpoints_only=endpoints_only, **kwds)



from numpy.testing import assert_array_equal
from pytest import approx

from landlab import RadialModelGrid


def test_radius_at_node():
    grid = RadialModelGrid(2, 8)
    assert grid.radius_at_node == approx(
        [
            2.0,
            2.0,
            2.0,
            2.0,
            2.0,
            1.0,
            2.0,
            2.0,
            1.0,
            1.0,
            2.0,
            1.0,
            0.0,
            1.0,
            2.0,
            1.0,
            1.0,
            2.0,
            2.0,
            1.0,
            2.0,
            2.0,
            2.0,
            2.0,
            2.0,
        ]
    )


def test_spacing_of_rings():
    grid = RadialModelGrid(2, spacing=1.0)
    assert grid.spacing_of_rings == approx(1.0)
    assert isinstance(grid.spacing_of_rings, float)

    grid = RadialModelGrid(2, spacing=11.0)
    assert grid.spacing_of_rings == approx(11.0)
    assert isinstance(grid.spacing_of_rings, float)


def test_number_of_rings():
    grid = RadialModelGrid(2)
    assert grid.number_of_rings == 2
    assert isinstance(grid.number_of_rings, int)

    grid = RadialModelGrid(4)
    assert grid.number_of_rings == 4


def test_nodes_per_ring():
    grid = RadialModelGrid(3, 8)
    assert_array_equal(grid.number_of_nodes_in_ring, [8, 16, 32])



import pytest

from landlab import RasterModelGrid
from landlab.plot.graph import _parse_locations_as_set, plot_graph


def _axes_arrows(ax):
    from matplotlib.patches import FancyArrow

    return [child for child in ax.get_children() if isinstance(child, FancyArrow)]


def test_parse_locations_from_str():
    assert _parse_locations_as_set("node") == {"node"}
    assert _parse_locations_as_set("node,link") == {"node", "link"}
    assert _parse_locations_as_set("node,link,cell,link") == {"node", "link", "cell"}
    assert _parse_locations_as_set("node  , link, patch  ") == {"node", "link", "patch"}


def test_parse_locations_from_iterable():
    assert _parse_locations_as_set(["cell"]) == {"cell"}
    assert _parse_locations_as_set(("patch", "corner")) == {"patch", "corner"}
    assert _parse_locations_as_set(("patch", "corner", "patch")) == {"patch", "corner"}
    assert _parse_locations_as_set((" patch  ", "corner  ")) == {"patch", "corner"}


def test_parse_locations_bad_value():
    with pytest.raises(ValueError, match=r"^unknown location "):
        _parse_locations_as_set("foo")
    with pytest.raises(ValueError, match=r"^unknown locations "):
        _parse_locations_as_set("cells,nodes")


@pytest.mark.parametrize("at", ["node", "link", "patch", "corner", "face", "cell"])
def test_plot_graph(at):
    grid = RasterModelGrid((3, 4))
    ax = plot_graph(grid, at=at)

    assert ax.get_ylim() == (-0.5, 2.5)
    assert ax.get_xlim() == (-0.5, 3.5)

    if at in ("patch", "cell"):
        assert len(ax.patches) == grid.number_of_elements(at)
    elif at in ("link", "face"):
        assert len(_axes_arrows(ax)) == grid.number_of_elements(at)
    else:
        assert len(ax.lines) == grid.number_of_elements(at)


def test_plot_graph_onto_existing():
    grid = RasterModelGrid((3, 4))
    ax = plot_graph(grid, at="node")
    plot_graph(grid, at="cell", axes=ax)

    assert (
        len(ax.patches) + len(ax.lines) == grid.number_of_nodes + grid.number_of_cells
    )



import numpy as np
import pytest
from hypothesis import given
from hypothesis.strategies import integers, lists
from numpy.testing import assert_array_almost_equal, assert_array_equal
from pytest import approx

from landlab.graph import TriGraph
from landlab.graph.hex.hex import (
    HorizontalHexTriGraph,
    HorizontalRectTriGraph,
    VerticalHexTriGraph,
    VerticalRectTriGraph,
)


def test_number_of_nodes_horizontal_rect():
    assert HorizontalRectTriGraph.number_of_nodes((1, 2)) == 2
    assert HorizontalRectTriGraph.number_of_nodes((1, 3)) == 3
    assert HorizontalRectTriGraph.number_of_nodes((2, 2)) == 4
    assert HorizontalRectTriGraph.number_of_nodes((2, 3)) == 6
    assert HorizontalRectTriGraph.number_of_nodes((3, 2)) == 6
    assert HorizontalRectTriGraph.number_of_nodes((3, 3)) == 9


def test_number_of_nodes_vertical_rect():
    assert VerticalRectTriGraph.number_of_nodes((1, 2)) == 2
    assert VerticalRectTriGraph.number_of_nodes((1, 3)) == 3
    assert VerticalRectTriGraph.number_of_nodes((2, 2)) == 4
    assert VerticalRectTriGraph.number_of_nodes((2, 3)) == 6
    assert VerticalRectTriGraph.number_of_nodes((3, 2)) == 6
    assert VerticalRectTriGraph.number_of_nodes((3, 3)) == 9


def test_number_of_nodes_horizontal_hex():
    assert HorizontalHexTriGraph.number_of_nodes((1, 2)) == 2
    assert HorizontalHexTriGraph.number_of_nodes((1, 3)) == 3
    assert HorizontalHexTriGraph.number_of_nodes((2, 2)) == 5
    assert HorizontalHexTriGraph.number_of_nodes((2, 3)) == 7
    assert HorizontalHexTriGraph.number_of_nodes((3, 2)) == 7
    assert HorizontalHexTriGraph.number_of_nodes((3, 3)) == 10


def test_number_of_nodes_vertical_hex():
    assert VerticalHexTriGraph.number_of_nodes((1, 2)) == 3
    assert VerticalHexTriGraph.number_of_nodes((1, 3)) == 4
    assert VerticalHexTriGraph.number_of_nodes((2, 2)) == 5
    assert VerticalHexTriGraph.number_of_nodes((2, 3)) == 7
    assert VerticalHexTriGraph.number_of_nodes((3, 2)) == 7
    assert VerticalHexTriGraph.number_of_nodes((3, 3)) == 10


@given(shape=lists(integers(min_value=3, max_value=1024), min_size=2, max_size=2))
def test_number_of_nodes_symetric_rect(shape):
    assert VerticalRectTriGraph.number_of_nodes(
        shape
    ) == HorizontalRectTriGraph.number_of_nodes(shape[::-1])


@given(shape=lists(integers(min_value=3, max_value=1024), min_size=2, max_size=2))
def test_number_of_nodes_symetric_hex(shape):
    assert VerticalHexTriGraph.number_of_nodes(
        shape
    ) == HorizontalHexTriGraph.number_of_nodes(shape[::-1])


@pytest.mark.parametrize("n_rows", (3,))
@pytest.mark.parametrize("node_layout", ("rect", "hex"))
@pytest.mark.parametrize("orientation", ("horizontal", "vertical"))
@pytest.mark.parametrize("at", ("nodes", "links", "patches"))
def test_create_hex_graph(n_rows, node_layout, orientation, at):
    expected = {
        "rect": {
            "horizontal": {"nodes": 6, "links": 9, "patches": 4},
            "vertical": {"nodes": 6, "links": 9, "patches": 4},
        },
        "hex": {
            "horizontal": {"nodes": 7, "links": 12, "patches": 6},
            "vertical": {"nodes": 7, "links": 12, "patches": 6},
        },
    }
    if orientation == "vertical":
        shape = (2, n_rows)
    else:
        shape = (n_rows, 2)
    graph = TriGraph(shape, node_layout=node_layout, orientation=orientation, sort=True)
    assert getattr(graph, f"number_of_{at}") == expected[node_layout][orientation][at]


def test_create_rect():
    """Test creating a hex graph with rectangular layout."""
    graph = TriGraph((3, 2), node_layout="rect", sort=True)

    assert graph.number_of_nodes == 6
    assert graph.number_of_links == 9
    assert graph.number_of_patches == 4


def test_create_hex():
    """Test creating a hex graph with hex layout."""
    graph = TriGraph((3, 2), node_layout="hex", sort=True)

    assert graph.number_of_nodes == 7
    assert graph.number_of_links == 12
    assert graph.number_of_patches == 6


@given(shape=lists(integers(min_value=3, max_value=32), min_size=2, max_size=2))
def test_spacing(shape):
    """Test spacing of nodes."""
    graph = TriGraph(shape)
    assert_array_almost_equal(graph.length_of_link, 1.0)

    graph = TriGraph(shape, spacing=2)
    assert_array_almost_equal(graph.length_of_link, 2.0)


@given(shape=lists(integers(min_value=3, max_value=32), min_size=2, max_size=2))
@pytest.mark.parametrize("orientation", ("horizontal", "vertical"))
@pytest.mark.parametrize("node_layout", ("hex", "rect"))
def test_origin_keyword(node_layout, orientation, shape):
    """Test setting the origin."""
    graph = TriGraph(shape)

    assert np.min(graph.x_of_node) == approx(0.0)
    assert np.min(graph.y_of_node) == approx(0.0)

    graph = TriGraph(shape, xy_of_lower_left=(0.5, 0.25))

    assert np.min(graph.x_of_node[0]) == approx(0.5)
    assert np.min(graph.y_of_node[0]) == approx(0.25)


def test_orientation():
    """Test vertical and horizontal orientation."""
    graph = TriGraph((3, 3), orientation="vertical")
    assert_array_almost_equal(
        graph.y_of_node, [0.0, 0.0, 0.5, 1.0, 1.0, 1.5, 2.0, 2.0, 2.5]
    )

    graph = TriGraph((3, 3), orientation="horizontal")
    assert_array_almost_equal(
        graph.x_of_node, [0.0, 1.0, 2.0, 0.5, 1.5, 2.5, 0.0, 1.0, 2.0]
    )


def test_perimeter_nodes_rect():
    graph = TriGraph((3, 4), node_layout="rect")
    assert_array_equal(graph.perimeter_nodes, [3, 7, 11, 10, 9, 8, 4, 0, 1, 2])


def test_perimeter_nodes_hex():
    graph = TriGraph((4, 2), node_layout="hex")
    assert_array_equal(graph.perimeter_nodes, [8, 11, 10, 9, 5, 2, 0, 1, 4])


def test_adjacent_nodes_at_node():
    graph = TriGraph((3, 3), node_layout="hex", sort=True)
    assert_array_equal(
        graph.adjacent_nodes_at_node,
        [
            [1, 4, 3, -1, -1, -1],
            [2, 5, 4, 0, -1, -1],
            [6, 5, 1, -1, -1, -1],
            [4, 7, 0, -1, -1, -1],
            [5, 8, 7, 3, 0, 1],
            [6, 9, 8, 4, 1, 2],
            [9, 5, 2, -1, -1, -1],
            [8, 3, 4, -1, -1, -1],
            [9, 7, 4, 5, -1, -1],
            [8, 5, 6, -1, -1, -1],
        ],
    )


def test_patches_at_node():
    grid = TriGraph((3, 3), node_layout="hex", sort=True)
    assert_array_equal(
        grid.patches_at_node,
        [
            [0, 2, -1, -1, -1, -1],
            [1, 3, 0, -1, -1, -1],
            [4, 1, -1, -1, -1, -1],
            [5, 2, -1, -1, -1, -1],
            [6, 8, 5, 2, 0, 3],
            [7, 9, 6, 3, 1, 4],
            [7, 4, -1, -1, -1, -1],
            [5, 8, -1, -1, -1, -1],
            [8, 6, 9, -1, -1, -1],
            [9, 7, -1, -1, -1, -1],
        ],
    )


@pytest.mark.parametrize("n_cols", (2, 3))
@pytest.mark.parametrize("n_rows", (2, 3))
def test_xy_of_node_rect_vertical(n_rows, n_cols):
    expected = {
        (2, 2): ([0, 1, 0, 1], [0, 0.5, 1, 1.5]),
        (2, 3): ([0, 2, 1, 0, 2, 1], [0, 0, 0.5, 1, 1, 1.5]),
        (3, 2): ([0, 1, 0, 1, 0, 1], [0, 0.5, 1, 1.5, 2, 2.5]),
        (3, 3): ([0, 2, 1, 0, 2, 1, 0, 2, 1], [0, 0, 0.5, 1, 1, 1.5, 2, 2, 2.5]),
    }
    x_of_node, y_of_node = VerticalRectTriGraph.xy_of_node((n_rows, n_cols))

    assert np.all(
        x_of_node / np.sin(np.pi / 3.0) == approx(expected[(n_rows, n_cols)][0])
    )
    assert np.all(y_of_node == approx(expected[(n_rows, n_cols)][1]))


@pytest.mark.parametrize("n_cols", (2, 3))
@pytest.mark.parametrize("n_rows", (1, 2, 3))
def test_xy_of_node_hex_vertical(n_rows, n_cols):
    expected = {
        (1, 2): ([1.0, 0, 1.0], [0, 0.5, 1]),
        (1, 3): ([1.0, 0, 2, 1.0], [0, 0.5, 0.5, 1]),
        (2, 2): ([1.0, 0, 1.0, 0, 1.0], [0, 0.5, 1, 1.5, 2]),
        (2, 3): ([1.0, 0, 2, 1.0, 0, 2, 1.0], [0, 0.5, 0.5, 1, 1.5, 1.5, 2]),
        (3, 2): ([1.0, 0, 1.0, 0, 1.0, 0.0, 1.0], [0, 0.5, 1, 1.5, 2, 2.5, 3]),
        (3, 3): (
            [1.0, 0, 2, 1.0, 0, 2, 1.0, 0, 2, 1.0],
            [0, 0.5, 0.5, 1, 1.5, 1.5, 2, 2.5, 2.5, 3],
        ),
    }
    x_of_node, y_of_node = VerticalHexTriGraph.xy_of_node((n_rows, n_cols))

    assert np.all(
        x_of_node / np.sin(np.pi / 3.0) == approx(expected[(n_rows, n_cols)][0])
    )
    assert np.all(y_of_node == approx(expected[(n_rows, n_cols)][1]))


def test_xy_of_node_spacing(hex_layout):
    x_of_node_expected, y_of_node_expected = hex_layout.xy_of_node((3, 4))
    x_of_node, y_of_node = hex_layout.xy_of_node((3, 4), spacing=2.0)

    assert_array_almost_equal(x_of_node / 2.0, x_of_node_expected)
    assert_array_almost_equal(y_of_node / 2.0, y_of_node_expected)


@pytest.mark.parametrize("n_cols", (2, 3))
@pytest.mark.parametrize("n_rows", (1, 2, 3))
def test_xy_of_node_lower_left(hex_layout, n_rows, n_cols):
    (x_of_node, y_of_node) = hex_layout.xy_of_node((n_rows, n_cols))

    assert np.min(x_of_node) == approx(0.0)
    assert np.min(y_of_node) == approx(0.0)



import numpy as np

from ..nodestatus import NodeStatus


class StatusGrid:
    def __init__(self, node_status):
        self._node_status = np.array(node_status)

    @property
    def node_status(self):
        """Status of nodes.

        Return an array of the status of a grid's nodes. The node status can
        be one of the following:
        - ``NodeStatus.CORE``
        - ``NodeStatus.FIXED_VALUE``
        - ``NodeStatus.FIXED_GRADIENT``
        - ``NodeStatus.LOOPED``
        - ``NodeStatus.CLOSED``
        """
        return self._node_status

    @node_status.setter
    def node_status(self, node_status):
        self._node_status = node_status

    def active_nodes(self):
        """Node IDs of all active (core & open boundary) nodes.

        core_nodes will return just core nodes.
        """
        (active_node_ids,) = np.where(self.node_status != NodeStatus.CLOSED)
        return active_node_ids

    def core_nodes(self):
        """Node IDs of all core nodes."""
        (core_node_ids,) = np.where(self.node_status == NodeStatus.CORE)
        return core_node_ids

    def boundary_nodes(self):
        """Node IDs of all boundary nodes."""
        (boundary_node_ids,) = np.where(self.node_status != NodeStatus.CORE)
        return boundary_node_ids

    def open_boundary_nodes(self):
        """Node id of all open boundary nodes."""
        (open_boundary_node_ids,) = np.where(
            (self.node_status != NodeStatus.CLOSED)
            & (self.node_status != NodeStatus.CORE)
        )
        return open_boundary_node_ids

    def closed_boundary_nodes(self):
        """Node id of all closed boundary nodes."""
        (closed_boundary_node_ids,) = np.where(self.node_status == NodeStatus.CLOSED)
        return closed_boundary_node_ids

    def fixed_gradient_boundary_nodes(self):
        """Node id of all fixed gradient boundary nodes."""
        (fixed_gradient_boundary_node_ids,) = np.where(
            self.node_status == NodeStatus.FIXED_GRADIENT
        )
        return fixed_gradient_boundary_node_ids

    def fixed_value_boundary_nodes(self):
        """Node id of all fixed value boundary nodes."""
        (fixed_value_boundary_node_ids,) = np.where(
            self.node_status == NodeStatus.FIXED_VALUE
        )
        return fixed_value_boundary_node_ids



import numpy as np

cimport cython
cimport numpy as np

from landlab.core.messages import warning_message

DTYPE_INT = int
ctypedef np.int_t DTYPE_INT_t

DTYPE_FLOAT = np.double
ctypedef np.double_t DTYPE_FLOAT_t


@cython.boundscheck(False)
cpdef find_lowest_node_on_lake_perimeter_c(
        np.ndarray[DTYPE_INT_t, ndim=2] node_nbrs,
        np.ndarray[DTYPE_INT_t, ndim=1] flood_status,
        np.ndarray[DTYPE_FLOAT_t, ndim=1] elev,
        np.ndarray[DTYPE_INT_t, ndim=1] nodes_this_depression,
        DTYPE_INT_t pit_count,
        DTYPE_FLOAT_t BIG_ELEV
    ):
    """Locate the lowest node on the margin of the "lake".

    Parameters
    ----------
    node_nbrs : (nnodes, 4) or (nnodes, 8) array of int
        The node neighbors, as stored by a DepressionFinderAndRouter
        component
    flood_status : nnodes array of int
        The node flooded status at the point of the function call, as stored
        by a DepressionFinderAndRouter component
    elev : nnodes array of float
        The elevations of each node in the grid
    nodes_this_depression : nnodes array of int
        Nodes that form a pit, followed by padding values to make up a nnodes-
        long array. This should be passed in with the first value as the
        pit node, then padding values, but it will be updated in place to
        reflect the nodes in the lake.
    pit_count : int
        The number of nodes currently in the lake.

    Returns
    -------
    (int, int)
        (Lowest node on the perimeter of a depression, updated pit_count)

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import FlowAccumulator, DepressionFinderAndRouter
    >>> from landlab.components.flow_routing.cfuncs import find_lowest_node_on_lake_perimeter_c
    >>> mg = RasterModelGrid((7, 7), xy_spacing=0.5)
    >>> z = mg.add_field('node', 'topographic__elevation', mg.node_x.copy())
    >>> z += 0.01 * mg.node_y
    >>> mg.at_node['topographic__elevation'].reshape(mg.shape)[2:5, 2:5] *= 0.1
    >>> fr = FlowAccumulator(mg, flow_director='D8')
    >>> fr.run_one_step()  # the flow "gets stuck" in the hole
    >>> df = DepressionFinderAndRouter(mg)

    >>> node_nbrs = df._node_nbrs
    >>> flood_status = df.flood_status
    >>> elev = df._elev
    >>> BIG_ELEV = df._BIG_ELEV
    >>> nodes_this_depression = mg.zeros('node', dtype=int)
    >>> nodes_this_depression[0] = 16
    >>> pit_count = 1

    >>> find_lowest_node_on_lake_perimeter_c(
    ...     node_nbrs, flood_status, elev, nodes_this_depression, pit_count,
    ...     BIG_ELEV
    ... )
    (23, 1)
    >>> nodes_this_depression[1] = 8
    >>> pit_count = 2
    >>> find_lowest_node_on_lake_perimeter_c(
    ...     node_nbrs, flood_status, elev, nodes_this_depression, pit_count,
    ...     BIG_ELEV
    ... )
    (0, 2)
    """
    # Start with the first node on the list, and an arbitrarily large elev
    cdef int lowest_node = nodes_this_depression[0]
    cdef DTYPE_FLOAT_t lowest_elev = BIG_ELEV

    # set up a worst-case scanario array for the pits, and a counter to pull
    # the good entries later:
    cdef int current_iter = 0

    # Codes for depression status
    cdef int _UNFLOODED = 0
    cdef int _PIT = 1
    cdef int _CURRENT_LAKE = 2
    cdef int _FLOODED = 3

    cdef int n
    cdef int nbr
    cdef int i

    while current_iter < pit_count:
        n = nodes_this_depression[current_iter]
        for nbr in node_nbrs[n]:
            if nbr != -1:
                if flood_status[nbr] == _UNFLOODED:
                    if elev[nbr] < lowest_elev:
                        lowest_node = nbr
                        lowest_elev = elev[nbr]
                elif (
                    flood_status[nbr] == _PIT
                    or flood_status[nbr] == _FLOODED
                ):
                    nodes_this_depression[pit_count] = nbr
                    pit_count += 1
                    flood_status[nbr] = _CURRENT_LAKE
        current_iter += 1
    if lowest_elev == BIG_ELEV:
        print("Unable to find drainage outlet for a lake.")
        print("In lake with " + str(len(nodes_this_depression)), "nodes:")
        print(str(nodes_this_depression))

        for i in nodes_this_depression:
            print("Node ID: ", i)
            print("Node Elevation: ", elev[i])
            print("Node Flood Status: ", flood_status[i])
            print("Node Neigbors: ", node_nbrs[i])
            print("Neighbor Elevations: ", elev[node_nbrs[i]])
            print("Neigbor Flood Status: ", flood_status[node_nbrs[i]])
        warning_message(
            """If you see no data values in any of the elevation terms
            this may because you have disconnected open nodes (which
            sometimes occurs during raster clipping.

            Consider running
            set_open_nodes_disconnected_from_watershed_to_closed
            which will remove isolated open nodes."""
        )
    return lowest_node, pit_count



import numpy as np
import pytest
from numpy.testing import assert_array_equal

from landlab import RasterModelGrid
from landlab.layers import EventLayers
from landlab.layers.eventlayers import _BlockSlice, _valid_keywords_or_raise


def test_EventLayersMixIn():
    grid = RasterModelGrid((4, 4))
    assert hasattr(grid, "event_layers")
    assert grid.event_layers.number_of_layers == 0
    assert grid.event_layers.number_of_stacks == 4


def test_setitem_with_scalar():
    layers = EventLayers(5)
    layers.add(1.0, age=3.0)
    layers.add(2.0, age=4.0)

    truth = np.array([[3.0, 3.0, 3.0, 3.0, 3.0], [4.0, 4.0, 4.0, 4.0, 4.0]])
    assert_array_equal(layers["age"], truth)

    layers["age"] = 2.0
    truth = np.array([[2.0, 2.0, 2.0, 2.0, 2.0], [2.0, 2.0, 2.0, 2.0, 2.0]])
    assert_array_equal(layers["age"], truth)


def test_set_item_with_1d():
    layers = EventLayers(5)
    layers.add(1.0, age=3.0)
    layers.add(2.0, age=4.0)

    truth = np.array([[3.0, 3.0, 3.0, 3.0, 3.0], [4.0, 4.0, 4.0, 4.0, 4.0]])
    assert_array_equal(layers["age"], truth)

    layers["age"] = [4.0, 7.0]

    truth = np.array([[4.0, 4.0, 4.0, 4.0, 4.0], [7.0, 7.0, 7.0, 7.0, 7.0]])
    assert_array_equal(layers["age"], truth)


def test_set_item_with_2d():
    layers = EventLayers(5)
    layers.add(1.0, age=3.0)
    layers.add(2.0, age=4.0)

    truth = np.array([[3.0, 3.0, 3.0, 3.0, 3.0], [4.0, 4.0, 4.0, 4.0, 4.0]])
    assert_array_equal(layers["age"], truth)

    layers["age"] = [[4.0, 4.0, 4.0, 4.0, 4.0], [7.0, 7.0, 7.0, 7.0, 7.0]]

    truth = np.array([[4.0, 4.0, 4.0, 4.0, 4.0], [7.0, 7.0, 7.0, 7.0, 7.0]])
    assert_array_equal(layers["age"], truth)


def test__str__():
    layers = EventLayers(5)
    layers.add(1.0, age=3.0)
    vals = str(layers)
    assert vals.splitlines() == [
        "number_of_layers: 1",
        "number_of_stacks: 5",
        "tracking: age",
    ]


def test__repr__():
    layers = EventLayers(5)
    layers.add(1.0, age=3.0)
    vals = repr(layers)
    assert vals == "EventLayers(5)"


def test_adding_untracked_layer():
    layers = EventLayers(3)
    layers.add(1.0, type=3.0, size="sand")
    layers.add([0.0, 0.0, 1.0], type=3.0, size="sand")
    with pytest.raises(ValueError):
        layers.add([1.0], type=3.0, size="sand", spam="eggs")


def test_reduce_with_no_args():
    layers = EventLayers(3)

    layers.add(1.5)
    layers.reduce()
    assert_array_equal(layers.dz, [[1.5, 1.5, 1.5]])

    layers.add(2.5)
    layers.reduce()
    assert_array_equal(layers.dz, [[4.0, 4.0, 4.0]])


def test_reduce_with_start():
    layers = EventLayers(3)

    layers.add([3.0, 4.0, 5.0])
    layers.add([1.0, 2.0, 0.5])
    layers.add([1.0, 2.0, 0.5])
    layers.reduce(1, 3)

    assert_array_equal(layers.dz, [[3.0, 4.0, 5.0], [2.0, 4.0, 1.0]])


def test_reduce_with_start_and_stop():
    layers = EventLayers(3)

    layers.add([3.0, 4.0, 5.0])
    layers.add([1.0, 2.0, 0.5])
    layers.add([1.0, 2.0, 0.5])
    layers.add([2.0, 5.0, 6.0])

    layers.reduce(1, 3)
    assert_array_equal(layers.dz, [[3.0, 4.0, 5.0], [2.0, 4.0, 1.0], [2.0, 5.0, 6.0]])


def test_reduce_with_start_and_stop_and_step():
    layers = EventLayers(3)

    layers.add([3.0, 4.0, 5.0])
    layers.add([1.0, 2.0, 0.5])
    layers.add([1.0, 2.0, 0.5])
    layers.add([2.0, 5.0, 6.0])

    layers.reduce(1, 3, 2)
    assert_array_equal(layers.dz, [[3.0, 4.0, 5.0], [2.0, 4.0, 1.0], [2.0, 5.0, 6.0]])


def test_reduce_with_stop_equals_start():
    layers = EventLayers(3)

    layers.add([3.0, 4.0, 5.0])
    layers.add([1.0, 2.0, 0.5])
    layers.add([1.0, 2.0, 0.5])
    layers.add([2.0, 5.0, 6.0])

    layers.reduce(1, 1)
    assert_array_equal(
        layers.dz, [[3.0, 4.0, 5.0], [1.0, 2.0, 0.5], [1.0, 2.0, 0.5], [2.0, 5.0, 6.0]]
    )


def test_reduce_with_stop_less_than_start():
    layers = EventLayers(3)

    layers.add([3.0, 4.0, 5.0])
    layers.add([1.0, 2.0, 0.5])
    layers.add([1.0, 2.0, 0.5])
    layers.add([2.0, 5.0, 6.0])

    with pytest.raises(ValueError):
        layers.reduce(3, 1)


def test_reduce_with_one_layer():
    layers = EventLayers(3)
    layers.add(1.0)
    layers.reduce()
    assert_array_equal(layers.dz, [[1, 1, 1]])


def test_reduce_with_no_layers():
    layers = EventLayers(3)
    layers.reduce()
    assert_array_equal(layers.dz, np.empty((0, 3)))


def test_reduce_with_attrs():
    layers = EventLayers(3)
    layers.add([1, 1, 1], age=0.0)
    layers.add([1, 2, 5], age=1.0)
    layers.add([2, 2, 2], age=2.0)
    layers.reduce(age=np.sum)
    assert_array_equal(layers["age"], [[3.0, 3.0, 3.0]])


def test_reduce_with_reducer():
    layers = EventLayers(3)
    layers.add([2, 2, 2], age=1.0)
    layers.add([2, 2, 2], age=3.0)
    layers.add([2, 2, 2], age=4.0)
    layers.reduce(1, 3, age=np.mean)
    assert_array_equal(layers["age"], [[1.0, 1.0, 1.0], [3.5, 3.5, 3.5]])


def test_reduce_with_start_too_big():
    layers = EventLayers(3)
    layers.add(1)
    layers.reduce(2, 4)
    assert_array_equal(layers.dz, [[1, 1, 1]])


def test_reduce_with_unknown_property():
    layers = EventLayers(3)
    for layer in range(4):
        layers.add(layer)
    with pytest.raises(TypeError):
        layers.reduce(1, 3, age=np.mean)


def test_reduce_with_missing_property():
    layers = EventLayers(3)
    for layer in range(4):
        layers.add(layer, age=layer)
    with pytest.raises(TypeError):
        layers.reduce(1, 3)


@pytest.mark.parametrize("args", [(), (4,), (0, 4), (0, 4, 4), (-4, 4, 4), (4, 0, -4)])
def test_reduce_args(args):
    layers = EventLayers(3)
    for layer in range(4):
        layers.add(layer)
    layers.reduce(4)
    assert_array_equal(layers.dz, [[6, 6, 6]])


def test_reduce_partial_block():
    layers = EventLayers(3)
    for layer in range(6):
        layers.add(layer)
    layers.reduce(1, 6, 4)
    assert_array_equal(layers.dz, [[0, 0, 0], [10, 10, 10], [5, 5, 5]])


def test_block_slice_no_args():
    block = _BlockSlice()
    assert block.start == 0
    assert block.stop is None
    assert block.step is None
    assert block.indices(4) == (0, 4, 4)


def test_block_slice_one_arg():
    block = _BlockSlice(4)
    assert block.start == 0
    assert block.stop == 4
    assert block.step is None
    assert block.indices(4) == (0, 4, 4)
    assert block.indices(3) == (0, 3, 3)


def test_block_slice_two_args():
    block = _BlockSlice(1, 4)
    assert block.start == 1
    assert block.stop == 4
    assert block.step is None
    assert block.indices(4) == (1, 4, 3)
    assert block.indices(5) == (1, 4, 3)
    assert block.indices(3) == (1, 3, 2)


def test_block_slice_three_args():
    block = _BlockSlice(1, 7, 2)
    assert block.start == 1
    assert block.stop == 7
    assert block.step == 2
    assert block.indices(8) == (1, 7, 2)

    block = _BlockSlice(1, 7, 7)
    assert block.start == 1
    assert block.stop == 7
    assert block.step == 6
    assert block.indices(8) == (1, 7, 6)


@pytest.mark.parametrize("args", [(0, 7, 2), (0, 8, 2), (0, 8, 21)])
@pytest.mark.parametrize("n_rows", [7, 8])
def test_block_slice_full_blocks(args, n_rows):
    start, stop, step = _BlockSlice(*args).indices(n_rows)
    assert (stop - start) % step == 0


def test_block_slice_stop_less_than_start():
    with pytest.raises(ValueError):
        _BlockSlice(5, 4)


def test_valid_keywords():
    _valid_keywords_or_raise({"foo": 0}, optional=["foo", "bar"])
    with pytest.raises(TypeError):
        _valid_keywords_or_raise({"foo": 0}, required=["foo", "bar"])
    with pytest.raises(TypeError):
        _valid_keywords_or_raise(["baz"], optional=["foo", "bar"])


@pytest.mark.parametrize("as_type", [list, tuple, set])
def test_valid_keywords_as_types(as_type):
    _valid_keywords_or_raise(as_type(["foo"]), optional=as_type(["foo", "bar"]))
    with pytest.raises(TypeError):
        _valid_keywords_or_raise(as_type(["foo"]), required=as_type(["foo", "bar"]))
    with pytest.raises(TypeError):
        _valid_keywords_or_raise(as_type(["baz"]), optional=as_type(["foo", "bar"]))


def test_valid_keywords_empty():
    _valid_keywords_or_raise([])
    _valid_keywords_or_raise([], optional=["foo"])
    with pytest.raises(TypeError):
        _valid_keywords_or_raise([], required=["foo"])



#!/usr/bin/env python
"""Tests for Record of SpeciesEvolver."""
from collections import OrderedDict

import numpy as np
import pandas as pd

from landlab.components.species_evolution.record import Record


def test_attributes():
    record = Record()
    np.testing.assert_equal(record.prior_time, np.nan)

    record.advance_time(10)

    record.set_value("vara", 1)
    record.set_value("varb", 2)

    np.testing.assert_equal(record.times, [0, 10])
    np.testing.assert_equal(record.prior_time, 0)
    np.testing.assert_equal(record.earliest_time, 0)
    np.testing.assert_equal(record.latest_time, 10)
    np.testing.assert_equal(record.variables, ["vara", "varb"])
    np.testing.assert_equal(record.count_of_time_steps, 2)
    np.testing.assert_equal(len(record), 2)

    df = pd.DataFrame({"time": [0, 10], "vara": [np.nan, 1], "varb": [np.nan, 2]})
    pd.testing.assert_frame_equal(record.data_frame, df, check_like=True)


def test_get_value():
    record = Record()
    record.set_value("vara", 1)
    record.advance_time(10)
    record.set_value("vara", 2)

    value = record.get_value("vara")
    np.testing.assert_equal(value, 2)

    value = record.get_value("vara", time=0)
    np.testing.assert_equal(value, 1)

    value = record.get_value("varb")
    np.testing.assert_equal(value, np.nan)


def test_set_value():
    record = Record()
    record.advance_time(10)

    record.set_value("vara", 1)
    d = OrderedDict([("time", [0, 10]), ("vara", [np.nan, 1])])
    np.testing.assert_equal(record._dict, d)

    record.set_value("vara", 3, time=0)
    d = OrderedDict([("time", [0, 10]), ("vara", [3, 1])])
    np.testing.assert_equal(record._dict, d)


def test_increment_value():
    record = Record()
    record.set_value("vara", 1)
    record.increment_value("vara", 100)

    d = OrderedDict([("time", [0]), ("vara", [101])])
    np.testing.assert_equal(record._dict, d)

    record.advance_time(10)
    record.increment_value("vara", 100)

    d = OrderedDict([("time", [0, 10]), ("vara", [101, 100])])
    np.testing.assert_equal(record._dict, d)

    record.increment_value("varb", 200, time=0)

    d = OrderedDict([("time", [0, 10]), ("vara", [101, 100]), ("varb", [200, np.nan])])
    np.testing.assert_equal(record._dict, d)


def test_initial_time():
    record = Record(initial_time=5)
    record.set_value("vara", 1)
    record.advance_time(10)

    d = OrderedDict([("time", [5, 15]), ("vara", [1, np.nan])])
    np.testing.assert_equal(record._dict, d)


def test_nonexistent_time():
    record = Record()
    record.set_value("vara", 1)

    np.testing.assert_raises(ValueError, record.get_value, "vara", 5)



import numpy as np

from ...core.utils import as_id_array
from ...graph.structured_quad.structured_quad import StructuredQuadGraphTopology
from . import _neighbors_at_link


def neighbors_at_link(shape, links):
    """Get neighbor links.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.components.overland_flow._links import neighbors_at_link

    >>> neighbors_at_link((3, 2), np.arange(7)) # doctest: +NORMALIZE_WHITESPACE
    array([[-1,  3, -1, -1],
           [ 2,  4, -1, -1], [-1,  5,  1, -1],
           [-1,  6, -1,  0],
           [ 5,  7, -1,  1], [-1, -1,  4,  2],
           [-1, -1, -1,  3]])
    """
    links = np.asarray(links, dtype=int)
    out = np.full((links.size, 4), -1, dtype=int)
    _neighbors_at_link.neighbors_at_link(links, shape, out)
    return out


def vertical_link_ids(shape):
    """Vertical links in a structured quad grid.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    (M, N) ndarray :
        Array of link IDs.

    Examples
    --------
    >>> from landlab.components.overland_flow._links import vertical_link_ids
    >>> vertical_link_ids((3, 4))
    array([[ 3,  4,  5,  6],
           [10, 11, 12, 13]])
    """
    layout = StructuredQuadGraphTopology(shape)
    return layout.vertical_links.reshape((shape[0] - 1, shape[1]))


def horizontal_link_ids(shape):
    """Horizontal links in a structured quad grid.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    (M, N) ndarray :
        Array of link IDs.

    Examples
    --------
    >>> from landlab.components.overland_flow._links import horizontal_link_ids
    >>> horizontal_link_ids((3, 4))
    array([[ 0,  1,  2],
           [ 7,  8,  9],
           [14, 15, 16]])
    """
    layout = StructuredQuadGraphTopology(shape)
    return layout.horizontal_links.reshape((shape[0], shape[1] - 1))


def vertical_south_link_neighbor(shape, vertical_ids, bad_index_value=-1):
    """Link IDs of south, vertical link neighbor.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    vertical_ids : array of int
        Array of all vertical link ids - MUST BE ARRAY OF LEN(VERTICAL_LINKS)
    bad_index_value: int, optional
        Value assigned to inactive indicies in the array.

    Returns
    -------
    ndarray :
        Link IDs of *south* vertical neighbor links. Length of
        number_of_vertical_links.

    Examples
    --------
    The following example uses this grid::

        *       *       *       *       *
        ^       ^       ^       ^       ^
       22      23      24      25      26
        |       |       |       |       |
        *       *       *       *       *
        ^       ^       ^       ^       ^
       13      14      15      16      17
        |       |       |       |       |
        *       *       *       *       *
        ^       ^       ^       ^       ^
        4       5       6       7       8
        |       |       |       |       |
        *       *       *       *       *

    .. note::

        Only vertical links are shown. When no neighbor is found,
        bad_index_value is returned.

        ``*`` indicates nodes

        Numeric values correspond to the vertical IDs.

    >>> from landlab import RasterModelGrid
    >>> from landlab.components.overland_flow._links import (
    ...     vertical_link_ids, vertical_south_link_neighbor
    ... )
    >>> rmg = RasterModelGrid((4, 5))
    >>> vertical_links = vertical_link_ids(rmg.shape)
    >>> vertical_south_link_neighbor(rmg.shape, vertical_links)
    ...     # doctest: +NORMALIZE_WHITESPACE
    array([-1, -1, -1, -1, -1,
            4,  5,  6,  7,  8,
            13, 14, 15, 16, 17])
    """
    vertical_links = StructuredQuadGraphTopology(shape).vertical_links
    vertical_links[shape[1] :] = vertical_links[: -shape[1]]
    vertical_links[: shape[1]] = bad_index_value

    return vertical_links


def vertical_west_link_neighbor(shape, vertical_ids, bad_index_value=-1):
    """Link IDs of west, vertical link neighbor.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    vertical_ids : array of int
        Array of all vertical link ids- MUST BE ARRAY OF LEN(VERTICAL_LINKS)
    bad_index_value: int, optional
        Value assigned to inactive indicies in the array.

    Returns
    -------
    ndarray :
        Link IDs of *west* vertical neighbor links. Length of
        number_of_vertical_links.

    Examples
    --------
    The following example uses this grid::

        *       *       *       *       *
        ^       ^       ^       ^       ^
       22      23      24      25      26
        |       |       |       |       |
        *       *       *       *       *
        ^       ^       ^       ^       ^
       13      14      15      16      17
        |       |       |       |       |
        *       *       *       *       *
        ^       ^       ^       ^       ^
        4       5       6       7       8
        |       |       |       |       |
        *       *       *       *       *

    .. note::

        Only vertical links are shown. When no neighbor is found,
        bad_index_value is returned.

        ``*`` indicates nodes

        Numeric values correspond to the vertical IDs.

    >>> from landlab import RasterModelGrid
    >>> from landlab.components.overland_flow._links import (
    ...     vertical_link_ids, vertical_west_link_neighbor
    ... )
    >>> rmg = RasterModelGrid((4, 5))
    >>> vertical_links = vertical_link_ids(rmg.shape)
    >>> vertical_west_link_neighbor(rmg.shape, vertical_links)
    ...     # doctest: +NORMALIZE_WHITESPACE
    array([-1,  4,  5,  6,  7,
           -1, 13, 14, 15, 16,
           -1, 22, 23, 24, 25])
    """
    vertical_links = StructuredQuadGraphTopology(shape).vertical_links.reshape(
        (shape[0] - 1, shape[1])
    )
    vertical_links[:, 1:] = vertical_links[:, :-1]
    vertical_links[:, 0] = bad_index_value

    return vertical_links.reshape(-1)


def vertical_north_link_neighbor(shape, vertical_ids, bad_index_value=-1):
    """Link IDs of north, vertical link neighbor.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    vertical_ids : array of int
        Array of all vertical link ids- MUST BE ARRAY OF LEN(VERTICAL_LINKS)
    bad_index_value: int, optional
        Value assigned to inactive indicies in the array.

    Returns
    -------
    ndarray :
        Link IDs of *north* vertical neighbor links. Length of
        number_of_vertical_links.

    Examples
    --------
    The following example uses this grid::

        *       *       *       *       *
        ^       ^       ^       ^       ^
       22      23      24      25      26
        |       |       |       |       |
        *       *       *       *       *
        ^       ^       ^       ^       ^
       13      14      15      16      17
        |       |       |       |       |
        *       *       *       *       *
        ^       ^       ^       ^       ^
        4       5       6       7       8
        |       |       |       |       |
        *       *       *       *       *

    .. note::

        Only vertical links are shown. When no neighbor is found,
        bad_index_value is returned.

        ``*`` indicates nodes

        Numeric values correspond to the vertical IDs.

    >>> from landlab import RasterModelGrid
    >>> from landlab.components.overland_flow._links import (
    ...     vertical_link_ids, vertical_north_link_neighbor
    ... )
    >>> rmg = RasterModelGrid((4, 5))
    >>> vertical_ids = vertical_link_ids(rmg.shape)
    >>> vertical_north_link_neighbor(rmg.shape, vertical_ids)
    ...     # doctest: +NORMALIZE_WHITESPACE
    array([13, 14, 15, 16, 17,
           22, 23, 24, 25, 26,
           -1, -1, -1, -1, -1])
    """
    vertical_links = StructuredQuadGraphTopology(shape).vertical_links
    vertical_links[: -shape[1]] = vertical_links[shape[1] :]
    vertical_links[-shape[1] :] = bad_index_value

    return vertical_links


def vertical_east_link_neighbor(shape, vertical_ids, bad_index_value=-1):
    """Link IDs of east, vertical link neighbor.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    vertical_ids : array of int
        Array of all vertical link ids - MUST BE ARRAY OF LEN(VERTICAL_LINKS)
    bad_index_value: int, optional
        Value assigned to inactive indicies in the array.

    Returns
    -------
    ndarray :
        Link IDs of *east* vertical neighbor links. Length of
        number_of_vertical_links.

    Examples
    --------
    The following example uses this grid::

        *       *       *       *       *
        ^       ^       ^       ^       ^
       22      23      24      25      26
        |       |       |       |       |
        *       *       *       *       *
        ^       ^       ^       ^       ^
       13      14      15      16      17
        |       |       |       |       |
        *       *       *       *       *
        ^       ^       ^       ^       ^
        4       5       6       7       8
        |       |       |       |       |
        *       *       *       *       *

    .. note::

        Only vertical links are shown. When no neighbor is found,
        bad_index_value is returned.

        ``*`` indicates nodes

        Numeric values correspond to the vertical IDs.

    >>> from landlab import RasterModelGrid
    >>> from landlab.components.overland_flow._links import (
    ...     vertical_link_ids, vertical_east_link_neighbor
    ... )
    >>> rmg = RasterModelGrid((4, 5))
    >>> vertical_links = vertical_link_ids(rmg.shape)
    >>> vertical_east_link_neighbor(rmg.shape, vertical_links)
    ...     # doctest: +NORMALIZE_WHITESPACE
    array([ 5,  6,  7,  8, -1,
           14, 15, 16, 17, -1,
           23, 24, 25, 26, -1])
    """
    vertical_links = StructuredQuadGraphTopology(shape).vertical_links.reshape(
        (shape[0] - 1, shape[1])
    )
    vertical_links[:, :-1] = vertical_links[:, 1:]
    vertical_links[:, -1] = bad_index_value

    return vertical_links.base


def active_link_ids(shape, node_status):
    """Get active links.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    node_status : array_link
        Status of nodes in grid.

    Returns
    -------
    ndarray :
        Links IDs at the active links.

    Examples
    --------
    >>> from landlab.grid import RasterModelGrid
    >>> from landlab.components.overland_flow._links import active_link_ids

    >>> rmg = RasterModelGrid((3, 4))
    >>> rmg.set_closed_boundaries_at_grid_edges(True, True, True, True)

    >>> status = rmg.status_at_node
    >>> status # doctest: +NORMALIZE_WHITESPACE
    array([4, 4, 4, 4,
           4, 0, 0, 4,
           4, 4, 4, 4], dtype=uint8)

    >>> active_link_ids((3, 4), status)
    array([8])
    """
    return as_id_array(np.where(is_active_link(shape, node_status))[0])


def is_active_link(shape, node_status):
    """Link IDs of active links.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    node_status : array_link
        Status of nodes in grid.

    Returns
    -------
    ndarray :
        Links IDs at the active links.

    Examples
    --------
    >>> from landlab.components.overland_flow._links import is_active_link
    >>> from landlab.grid.nodestatus import NodeStatus

    >>> status = [
    ...     [NodeStatus.CLOSED, NodeStatus.CLOSED, NodeStatus.CLOSED],
    ...     [NodeStatus.CLOSED, NodeStatus.CORE, NodeStatus.CLOSED],
    ...     [NodeStatus.CLOSED, NodeStatus.CORE, NodeStatus.CLOSED],
    ...     [NodeStatus.CLOSED, NodeStatus.CLOSED, NodeStatus.CLOSED],
    ... ]
    >>> is_active_link((4, 3), status) # doctest: +NORMALIZE_WHITESPACE
    array([False, False,
           False, False, False,
           False, False,
           False, True, False,
           False, False,
           False, False, False,
           False, False], dtype=bool)
    """
    from ...grid.linkstatus import is_active_link

    node_status = np.asarray(node_status).reshape(-1)

    if np.prod(shape) != node_status.size:
        raise ValueError(
            "node status array does not match size of grid "
            "(%d != %d)" % (np.prod(shape), len(node_status))
        )

    # status_at_link_start = node_status.flat[node_id_at_link_start(shape)]
    # status_at_link_end = node_status.flat[node_id_at_link_end(shape)]

    # status_at_link = node_status[StructuredQuadGraphTopology(shape).nodes_at_link]

    return is_active_link(node_status[StructuredQuadGraphTopology(shape).nodes_at_link])


def vertical_active_link_ids(shape, active_ids, bad_index_value=-1):
    """ID of vertical active links.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    active_ids : array of int
        Array of all active link ids
    bad_index_value: int, optional
        Value assigned to inactive indicies in the array.

    Returns
    -------
    ndarray :
        Link IDs at the VERTICAL active links. Length of
        number_of_vertical_links.

    Examples
    --------
    The following example uses this grid::

          *---I-->*---I-->*---I-->*---I-->*
          ^       ^       ^       ^       ^
          I       I       I       I       I
          |       |       |       |       |
          *---I-->o---H-->o---H-->o---I-->*
          ^       ^       ^       ^       ^
          I       6       7       8       I
          |       |       |       |       |
          *---I-->o---H-->o---H-->o---I-->*
          ^       ^       ^       ^       ^
          I       I       I       I       I
          |       |       |       |       |
          *---I-->*---I-->*---I-->*---I-->*

    .. note::

        ``*`` indicates the nodes that are set to `NodeStatus.CLOSED`

        ``o`` indicates the nodes that are set to `NodeStatus.CORE`

        ``I`` indicates the links that are set to `LinkStatus.INACTIVE`

        ``H`` indicates horizontal active ids, which are ignored by this
        function

        Numeric values correspond to the vertical `LinkStatus.ACTIVE` IDs.

    >>> from landlab import RasterModelGrid
    >>> from landlab.components.overland_flow._links import (active_link_ids,
    ...     vertical_active_link_ids)

    >>> rmg = RasterModelGrid((4, 5))
    >>> active_ids = active_link_ids((4, 5), rmg.status_at_node)
    >>> active_ids # doctest: +NORMALIZE_WHITESPACE
    array([ 5,  6,  7,
            9, 10, 11, 12,
           14, 15, 16,
           18, 19, 20, 21,
           23, 24, 25])

    >>> vertical_active_link_ids((4, 5), active_ids)
    ...     # doctest: +NORMALIZE_WHITESPACE
    array([-1,  5,  6,  7, -1,
           -1, 14, 15, 16, -1,
           -1, 23, 24, 25, -1])

    >>> rmg.set_closed_boundaries_at_grid_edges(True, True, True, True)
    >>> status = rmg.status_at_node
    >>> active_ids = active_link_ids((4, 5), status)
    >>> vertical_active_link_ids((4, 5), active_ids)
    ...     # doctest: +NORMALIZE_WHITESPACE
    array([-1, -1, -1, -1, -1,
           -1, 14, 15, 16, -1,
           -1, -1, -1, -1, -1])
    """
    number_of_vertical_links = (shape[0] - 1) * shape[1]
    out = np.full(number_of_vertical_links, bad_index_value, dtype=int)
    vertical_ids = active_ids[np.where(is_vertical_link(shape, active_ids))]

    out[nth_vertical_link(shape, vertical_ids)] = vertical_ids
    return out


def _number_of_links(shape):
    """Number of links in a structured quad grid.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    int :
        Number of links in grid.

    Examples
    --------
    >>> from landlab.components.overland_flow._links import _number_of_links
    >>> _number_of_links((3, 4))
    17
    """
    return (shape[0] - 1) * shape[1] + shape[0] * (shape[1] - 1)
    # return number_of_vertical_links(shape) + number_of_horizontal_links(shape)


def number_of_vertical_links(shape):
    """Number of vertical links in a structured quad grid.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    int :
        Number of vertical links in grid.

    Examples
    --------
    >>> from landlab.components.overland_flow._links import number_of_vertical_links
    >>> number_of_vertical_links((3, 4))
    8
    """
    return (shape[0] - 1) * shape[1]


def number_of_horizontal_links(shape):
    """Number of horizontal links in a structured quad grid.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.

    Returns
    -------
    int :
        Number of horizontal links in grid.

    Examples
    --------
    >>> from landlab.components.overland_flow._links import number_of_horizontal_links
    >>> number_of_horizontal_links((3, 4))
    9
    """
    return shape[0] * (shape[1] - 1)


def is_vertical_link(shape, links):
    """Test if links are vertical.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    links : array of int
        Array of link ids to test.

    Returns
    -------
    ndarray of bool
        `True` for links that are vertical.

    Examples
    --------
    >>> from landlab.components.overland_flow._links import (is_vertical_link,
    ...     _number_of_links)
    >>> import numpy as np
    >>> shape = (3, 4)
    >>> links = np.arange(_number_of_links(shape))
    >>> is_vertical_link(shape, links) # doctest: +NORMALIZE_WHITESPACE
    array([False, False, False,  True,  True,  True,  True,
           False, False, False,  True,  True,  True,  True,
           False, False, False], dtype=bool)
    """
    return ((links % (2 * shape[1] - 1)) >= shape[1] - 1) & (
        links < _number_of_links(shape)
    )


def nth_vertical_link(shape, links):
    """Convert link ID to vertical link ID.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    links : array of int
        Array of link ids to test.

    Returns
    -------
    ndarray of int
        The link ID as the nth vertical links.

    Examples
    --------
    >>> from landlab.components.overland_flow._links import nth_vertical_link
    >>> shape = (3, 4)
    >>> nth_vertical_link(shape, 4)
    1
    >>> nth_vertical_link(shape, (3, 4, 11))
    array([0, 1, 5])
    """
    links = np.asarray(links, dtype=int)
    return as_id_array(
        (links // (2 * shape[1] - 1)) * shape[1]
        + links % (2 * shape[1] - 1)
        - (shape[1] - 1)
    )


def horizontal_active_link_ids(shape, active_ids, bad_index_value=-1):
    """ID of horizontal active links.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    active_ids : array of int
        Array of all active link ids
    bad_index_value: int, optional
        Value assigned to inactive indicies in the array.

    Returns
    -------
    ndarray :
        Link IDs at the HORIZONTAL active links. Length of
        number_of_horizontal_links.

    Examples
    --------

    The following example uses this grid::

          *---I-->*---I-->*---I-->*---I-->*
          ^       ^       ^       ^       ^
          I       I       I       I       I
          |       |       |       |       |
          *---I-->o--24-->o--25-->o---I-->*
          ^       ^       ^       ^       ^
          I       V       V       V       I
          |       |       |       |       |
          *---I-->o--20-->o--21-->o---I-->*
          ^       ^       ^       ^       ^
          I       I       I       I       I
          |       |       |       |       |
          *---I-->*---I-->*---I-->*---I-->*

    .. note::

        ``*`` indicates the nodes that are set to `NodeStatus.CLOSED`

        ``o`` indicates the nodes that are set to `NodeStatus.CORE`

        ``I`` indicates the links that are set to `LinkStatus.INACTIVE`

        ``V`` indicates vertical active ids, which are ignored by this
        function.

        Numeric values correspond to the horizontal `LinkStatus.ACTIVE`  ID.

    >>> from landlab import RasterModelGrid
    >>> from landlab.components.overland_flow._links import (active_link_ids,
    ...     horizontal_active_link_ids)

    >>> rmg = RasterModelGrid((4, 5))
    >>> rmg.set_closed_boundaries_at_grid_edges(True, True, True, True)

    >>> status = rmg.status_at_node
    >>> status # doctest: +NORMALIZE_WHITESPACE
    array([4, 4, 4, 4, 4,
           4, 0, 0, 0, 4,
           4, 0, 0, 0, 4,
           4, 4, 4, 4, 4], dtype=uint8)
    >>> active_ids = active_link_ids((4,5), status)

    >>> horizontal_active_link_ids((4,5), active_ids)
    ...     # doctest: +NORMALIZE_WHITESPACE
    array([-1, -1, -1, -1,
           -1, 10, 11, -1,
           -1, 19, 20, -1,
           -1, -1, -1, -1])
    """
    number_of_horizontal_links = shape[0] * (shape[1] - 1)
    out = np.full(number_of_horizontal_links, bad_index_value, dtype=int)
    horizontal_ids = active_ids[np.where(~is_vertical_link(shape, active_ids))]

    out[nth_horizontal_link(shape, horizontal_ids)] = horizontal_ids
    return out


def nth_horizontal_link(shape, links):
    """Convert link ID to horizontal link ID.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    links : array of int
        Array of link ids to test.

    Returns
    -------
    ndarray of int
        The link ID as the nth horizontal links.

    Examples
    --------
    >>> from landlab.components.overland_flow._links import nth_horizontal_link
    >>> shape = (3, 4)
    >>> nth_horizontal_link(shape, 16)
    8
    >>> nth_horizontal_link(shape, (1, 7, 8))
    array([1, 3, 4])
    """
    links = np.asarray(links, dtype=int)
    return as_id_array(
        (links // (2 * shape[1] - 1)) * (shape[1] - 1) + links % (2 * shape[1] - 1)
    )


def is_horizontal_link(shape, links):
    """Test if a link is horizontal.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    links : array of int
        Array of link ids to test.

    Returns
    -------
    ndarray of bool
        `True` for links that are horizontal.

    Examples
    --------
    >>> from landlab.components.overland_flow._links import (is_horizontal_link,
    ...     _number_of_links)
    >>> import numpy as np
    >>> shape = (3, 4)
    >>> links = np.arange(_number_of_links(shape))
    >>> is_horizontal_link(shape, links) # doctest: +NORMALIZE_WHITESPACE
    array([ True,  True,  True, False, False, False, False,
            True,  True,  True, False, False, False, False,
            True,  True,  True], dtype=bool)
    """
    return (~is_vertical_link(shape, links)) & (links < _number_of_links(shape))


def horizontal_west_link_neighbor(shape, horizontal_ids, bad_index_value=-1):
    """ID of west, horizontal link neighbor.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    horizontal_ids : array of int
        Array of all horizontal link ids - *must be of len(horizontal_links)*
    bad_index_value: int, optional
        Value assigned to inactive indicies in the array.

    Returns
    -------
    ndarray :
        Link IDs of west horizontal neighbor links. Length of
        number_of_horizontal_links.

    Examples
    --------
    The following example uses this grid::

        *--27-->*--28-->*--29-->*--30-->*



        *--18-->*--19-->*--20-->*--21-->*



        *---9-->*--10-->*--11-->*--12-->*



        *---0-->*---1-->*---2-->*---3-->*

    .. note::

        Only horizontal links are shown. When no neighbor is found,
        bad_index_value is returned.

        ``*`` indicates nodes

        Numeric values correspond to the horizontal IDs.

    >>> from landlab import RasterModelGrid
    >>> from landlab.components.overland_flow._links import (horizontal_link_ids,
    ...     horizontal_west_link_neighbor)
    >>> rmg = RasterModelGrid((4, 5))
    >>> horizontal_links = horizontal_link_ids(rmg.shape).flatten()
    >>> horizontal_west_link_neighbor(rmg.shape, horizontal_links)
    array([-1,  0,  1,  2, -1,  9, 10, 11, -1, 18, 19, 20, -1, 27, 28, 29])
    """
    links = np.roll(horizontal_ids.reshape((shape[0], shape[1] - 1)), 1, axis=1)
    links[:, 0] = bad_index_value

    return links.reshape(-1)


def horizontal_east_link_neighbor(shape, horizontal_ids, bad_index_value=-1):
    """IDs of east, horizontal link neighbor.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    horizontal_ids : array of int
        Array of all horizontal link ids - *must be of len(horizontal_links)*
    bad_index_value: int, optional
        Value assigned to inactive indicies in the array.

    Returns
    -------
    ndarray :
        Link IDs of east horizontal neighbor links. Length of
        number_of_horizontal_links.

    Examples
    --------
    The following example uses this grid::

        *--27-->*--28-->*--29-->*--30-->*



        *--18-->*--19-->*--20-->*--21-->*



        *---9-->*--10-->*--11-->*--12-->*



        *---0-->*---1-->*---2-->*---3-->*

    .. note::

        Only horizontal links are shown. When no neighbor is found,
        bad_index_value is returned.

        ``*`` indicates nodes

        Numeric values correspond to the horizontal `LinkStatus.ACTIVE` IDs.

    >>> from landlab import RasterModelGrid
    >>> from landlab.components.overland_flow._links import (horizontal_link_ids,
    ...     horizontal_east_link_neighbor)
    >>> rmg = RasterModelGrid((4, 5))
    >>> horizontal_links = horizontal_link_ids(rmg.shape).flatten()
    >>> horizontal_east_link_neighbor(rmg.shape, horizontal_links)
    array([ 1,  2,  3, -1, 10, 11, 12, -1, 19, 20, 21, -1, 28, 29, 30, -1])
    """
    links = np.roll(horizontal_ids.reshape((shape[0], shape[1] - 1)), -1, axis=1)
    links[:, -1] = -1

    return links.reshape(-1)


def horizontal_north_link_neighbor(shape, horizontal_ids, bad_index_value=-1):
    """ID of north, horizontal link neighbor.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    horizontal_ids : array of int
        Array of all horizontal link ids - *must be of len(horizontal_links)*
    bad_index_value: int, optional
        Value assigned to inactive indicies in the array.

    Returns
    -------
    ndarray :
        Link IDs of north horizontal neighbor links. Length of
        number_of_horizontal_links.

    Examples
    --------
    The following example uses this grid::

        *--27-->*--28-->*--29-->*--30-->*



        *--18-->*--19-->*--20-->*--21-->*



        *---9-->*--10-->*--11-->*--12-->*



        *---0-->*---1-->*---2-->*---3-->*

    .. note::

        Only horizontal links are shown. When no neighbor is found,
        bad_index_value is returned.

        ``*`` indicates nodes

        Numeric values correspond to the horizontal `LinkStatus.ACTIVE` IDs.

    >>> from landlab import RasterModelGrid
    >>> from landlab.components.overland_flow._links import (horizontal_link_ids,
    ...     horizontal_north_link_neighbor)
    >>> rmg = RasterModelGrid((4, 5))
    >>> horizontal_links = horizontal_link_ids(rmg.shape).flatten()
    >>> horizontal_north_link_neighbor(rmg.shape, horizontal_links)
    array([ 9, 10, 11, 12, 18, 19, 20, 21, 27, 28, 29, 30, -1, -1, -1, -1])
    """
    links = np.roll(horizontal_ids.reshape((shape[0], shape[1] - 1)), -1, axis=0)
    links[-1, :] = bad_index_value

    return links.reshape(-1)


def horizontal_south_link_neighbor(shape, horizontal_ids, bad_index_value=-1):
    """ID of south horizontal link neighbor.

    Parameters
    ----------
    shape : tuple of int
        Shape of grid of nodes.
    horizontal_ids : array of int
        Array of all horizontal link ids *must be of len(horizontal_links)*.
    bad_index_value: int, optional
        Value assigned to inactive indicies in the array.

    Returns
    -------
    ndarray :
        Link IDs of south horizontal neighbor links. Length of
        number_of_horizontal_links.

    Examples
    --------
    The following example uses this grid::

        *--27-->*--28-->*--29-->*--30-->*



        *--18-->*--19-->*--20-->*--21-->*



        *---9-->*--10-->*--11-->*--12-->*



        *---0-->*---1-->*---2-->*---3-->*

    .. note::

        Only horizontal links are shown. When no neighbor is found,
        bad_index_value is returned.

        ``*`` indicates nodes

        Numeric values correspond to the horizontal IDs.

    >>> from landlab import RasterModelGrid
    >>> from landlab.components.overland_flow._links import (horizontal_link_ids,
    ...     horizontal_north_link_neighbor)
    >>> rmg = RasterModelGrid((4, 5))
    >>> horizontal_links = horizontal_link_ids(rmg.shape).flatten()
    >>> horizontal_south_link_neighbor(rmg.shape, horizontal_links)
    array([-1, -1, -1, -1,  0,  1,  2,  3,  9, 10, 11, 12, 18, 19, 20, 21])
    """
    links = np.roll(horizontal_ids.reshape((shape[0], shape[1] - 1)), 1, axis=0)
    links[0, :] = bad_index_value

    return links.reshape(-1)



import re
from collections import OrderedDict


class GraphConvention:

    """Define a naming convention for graph elements."""

    def __init__(self, node, edge, face, nodes=None, edges=None, faces=None):
        """Define a graph element naming convention.

        Parameters
        ----------
        node : str
            Name to use for "nodes".
        edge : str
            Name to use for "edges".
        face : str
            Name to use for "faces".
        nodes : str, optional
            Plural of "nodes"
        edges : str, optional
            Plural of "edges"
        faces : str, optional
            Plural of "faces"

        Examples
        --------
        >>> from landlab.graph import GraphConvention
        >>> convention = GraphConvention("node", "edge", "face")
        >>> convention.node
        'node'
        >>> convention.edge
        'edge'
        >>> convention.edges
        'edges'

        >>> convention = GraphConvention("node", "link", "patch", faces="patches")
        >>> convention.face
        'patch'
        >>> convention.faces
        'patches'
        """
        self._node = node
        self._edge = edge
        self._face = face
        self._nodes = nodes or node + "s"
        self._edges = edges or edge + "s"
        self._faces = faces or face + "s"

    @property
    def node(self):
        """The name of things that are points."""
        return self._node

    @property
    def nodes(self):
        """The plural name for node."""
        return self._nodes

    @property
    def edge(self):
        """The name of lines that connect two nodes."""
        return self._edge

    @property
    def edges(self):
        """The plural name of edge."""
        return self._edges

    @property
    def face(self):
        """The name of objects made up of a closed set of edges."""
        return self._face

    @property
    def faces(self):
        """The plural name for face."""
        return self._faces


class ConventionConverter:

    """Convert between graph element naming conventions."""

    CONVENTION = {
        "nlp": GraphConvention("node", "link", "patch", faces="patches"),
        "nef": GraphConvention("node", "edge", "face"),
        "cfc": GraphConvention("corner", "face", "cell"),
    }

    def __init__(self, convention="nlp"):
        """

        Parameters
        ----------
        convention : {"nlp", "nef", "cfc"}
            Naming convention to which to conform.
        """
        self.convention = convention

    @property
    def convention(self):
        """Name of the convention to conform to."""
        return self._convention

    @convention.setter
    def convention(self, val):
        if val not in ("nlp", "cfc", "nef"):
            raise ValueError(f"convention not understood ({val})")
        self._convention = val

    @staticmethod
    def _convention_mapper(from_convention, to_convention):
        """Create a mapper to convert names between conventions."""
        elements = ("nodes", "edges", "faces", "node", "edge", "face")

        from_elements = [getattr(from_convention, name) for name in elements]
        to_elements = [getattr(to_convention, name) for name in elements]

        return OrderedDict(
            list(zip(to_elements, from_elements))
            + list(zip(from_elements, to_elements))
        )

    @classmethod
    def _as_convention(cls, name):
        """Get a named convention as a GraphConvention object."""
        try:
            return cls.CONVENTION[name]
        except KeyError as exc:
            raise ValueError(f"convention not understood ({name})") from exc

    def conform(self, name, from_convention):
        """Convert a name to a new convention.

        Parameters
        ----------
        name : str
            A name in the given convention.
        from_convention : {"nlp", "nef", "cfc"}
            Naming convention of the provided string.

        Returns
        -------
        str
            The name converted to the convention.

        Examples
        --------
        >>> from landlab.graph import ConventionConverter

        >>> converter = ConventionConverter("nlp")

        >> converter.conform("nodes_at_edge", "nef")
        'nodes_at_link'

        >>> converter.conform("node_at_cell", "cfc")
        'corner_at_patch'

        >>> converter = ConventionConverter("nef")
        >>> converter.conform("faces_at_cell", "cfc")
        'edges_at_face'
        """
        if from_convention is self.convention:
            return name

        mapping = ConventionConverter._convention_mapper(
            ConventionConverter._as_convention(from_convention),
            ConventionConverter._as_convention(self.convention),
        )

        def rename_words(m):
            key = m.group(0)
            return mapping[key]

        pattern = "|".join(mapping.keys())
        return re.sub(pattern, rename_words, name)



"""Get the landlab tutorial and teaching notebooks.

Run this script to fetch the set of *landlab* notebooks compatible with
an installed version of landlab.

Usage
-----

Get notebooks for a currently installed *landlab*,

    $ python -m notebooks

To get notebooks for a particular version of *landlab*, provide
a version number as an argument. For example,

    $ python -m notebooks 2.5.0
"""
import argparse
import os
import pathlib
import sys
import tarfile
from urllib.error import HTTPError
from urllib.parse import urljoin
from urllib.request import urlopen

from packaging.version import Version


def main(version=None):
    if version in ("master", "latest", "dev"):
        version = None

    if not version or (version := Version(version)).is_devrelease:
        tag = "master"
    else:
        tag = "v" + version.base_version

    notebooks = NotebookFetcher(tag)

    out(f"fetching notebooks for landlab {notebooks.version}")
    out(f"{notebooks.url}")

    try:
        stream = notebooks.open()
    except NotebookError as error:
        err(str(error))
        return -1

    with tarfile.open(fileobj=stream, mode="r|gz") as tfile:
        base = NotebookExtractor(tfile).extract()

    out(f"notebooks have been extracted into {base}")
    out("To run the notebooks first install the required dependencies:")
    out("")
    out(f"    $ conda install --file={base}/requirements-notebooks.txt")
    out("")
    out("and then open the welcome notebook:")
    out("")
    out(f"    $ jupyter notebook {base}/notebooks/welcome.ipynb")
    print(base)

    return 0


class NotebookError(Exception):
    def __init__(self, msg):
        self._msg = msg

    def __str__(self):
        return self._msg


class NotebookFetcher:
    URL = "https://github.com/landlab/landlab/archive/refs"

    def __init__(self, version):
        self._version = version

    @property
    def version(self):
        return self._version

    @property
    def url(self):
        return urljoin(NotebookFetcher.URL, f"{self.version}.tar.gz")

    def open(self):
        try:
            stream = urlopen(self.url)
        except HTTPError as error:
            if error.code == 404:
                msg = f"unable to find notebooks for requested landlab version ({self.version})"
            else:
                msg = f"unable to fetch notebooks ({error.reason})"
            raise NotebookError(msg) from error
        else:
            return stream


class NotebookExtractor:
    def __init__(self, tfile):
        self._tfile = tfile
        self._names = []

    def extract(self):
        self._tfile.extractall(members=self._notebooks())
        return self.base

    def _notebooks(self):
        for tarinfo in self._tfile:
            parts = pathlib.Path(tarinfo.name).parts
            if (len(parts) > 1 and parts[1] == "notebooks") or (
                parts[-1] == "requirements-notebooks.txt"
            ):
                self._names.append(tarinfo.name)
                yield tarinfo

    @property
    def names(self):
        return sorted(self._names)

    @property
    def base(self):
        return pathlib.Path(os.path.commonprefix(self.names))


def out(text):
    print("\033[1m" + text + "\033[0m", file=sys.stderr)


def err(text):
    print("\033[91m" + text + "\033[0m", file=sys.stderr)


if __name__ == "__main__":
    try:
        import landlab
    except ModuleNotFoundError:
        version = ""
    else:
        version = landlab.__version__

    parser = argparse.ArgumentParser(
        description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument(
        "version",
        metavar="VERSION",
        nargs="?",
        default=version,
        help="a landlab version (e.g. 2.5.0)",
    )
    args = parser.parse_args()

    sys.exit(main(args.version))



"""
Created on Wed Jan 13 15:34:08 2021

@author: laure
"""

import numpy as np

from landlab import HexModelGrid, RadialModelGrid, RasterModelGrid, VoronoiDelaunayGrid
from landlab.utils.window_statistic import calculate_window_statistic


def test_hex_grid():
    grid = HexModelGrid((5, 4), spacing=10)
    grid.status_at_node[grid.status_at_node == 1] = grid.BC_NODE_IS_CLOSED
    z = grid.add_zeros("topographic__elevation", at="node")
    z += np.arange(len(z))
    relief = calculate_window_statistic(
        grid,
        "topographic__elevation",
        np.ptp,
        search_radius=15,
        calc_on_closed_nodes=False,
    )

    check = [
        np.nan,
        np.nan,
        np.nan,
        np.nan,
        np.nan,
        6.0,
        7.0,
        7.0,
        np.nan,
        np.nan,
        11.0,
        12.0,
        12.0,
        11.0,
        np.nan,
        np.nan,
        7.0,
        7.0,
        6.0,
        np.nan,
        np.nan,
        np.nan,
        np.nan,
        np.nan,
    ]

    np.testing.assert_equal(relief, check)


def test_calc_with_interior_closed_nodes():
    grid = RasterModelGrid((5, 5), xy_spacing=10.0)
    grid.status_at_node[[7, 8, 12, 13]] = grid.BC_NODE_IS_CLOSED
    z = grid.add_zeros("topographic__elevation", at="node")
    z += np.arange(len(z))
    relief = calculate_window_statistic(
        grid,
        "topographic__elevation",
        np.ptp,
        search_radius=15,
        calc_on_closed_nodes=False,
    )

    check = [
        6.0,
        6.0,
        5.0,
        7.0,
        6.0,
        11.0,
        11.0,
        np.nan,
        np.nan,
        11.0,
        11.0,
        12.0,
        np.nan,
        np.nan,
        10.0,
        11.0,
        12.0,
        12.0,
        10.0,
        10.0,
        6.0,
        7.0,
        7.0,
        7.0,
        6.0,
    ]

    np.testing.assert_equal(relief, check)


def test_voronoi_grid():
    x = [0, 0.001, 0.002, 0.003, 1, 1.001, 1.002, 1.003, 2, 2.001, 2.002, 2.003]
    y = [0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3]
    grid = VoronoiDelaunayGrid(x, y)
    z = grid.add_zeros("topographic__elevation", at="node")
    z += np.arange(len(z))
    relief = calculate_window_statistic(
        grid, "topographic__elevation", np.ptp, search_radius=1.5
    )

    check = [4.0, 5.0, 4.0, 7.0, 8.0, 7.0, 7.0, 8.0, 7.0, 4.0, 5.0, 4.0]

    np.testing.assert_equal(relief, check)


def test_radial_grid():
    grid = RadialModelGrid(2, 6, 1)
    grid.status_at_node[grid.status_at_node == 1] = grid.BC_NODE_IS_CLOSED
    z = grid.add_zeros("topographic__elevation", at="node")
    z += np.arange(len(z))
    relief = calculate_window_statistic(
        grid,
        "topographic__elevation",
        np.ptp,
        search_radius=1.5,
        calc_on_closed_nodes=False,
    )

    check = [
        np.nan,
        np.nan,
        np.nan,
        np.nan,
        np.nan,
        4.0,
        5.0,
        np.nan,
        7.0,
        8.0,
        7.0,
        np.nan,
        5.0,
        4.0,
        np.nan,
        np.nan,
        np.nan,
        np.nan,
        np.nan,
    ]

    np.testing.assert_equal(relief, check)



import numpy as np
import pytest
from numpy import testing

from landlab import RasterModelGrid
from landlab.components import ErosionDeposition, FlowAccumulator


def test_Ff_too_high_vals():
    """
    Test that instantiating ErosionDeposition with a F_f value > 1 throws a
    ValueError.
    """

    # set up a 5x5 grid with one open outlet node and low initial elevations.
    nr = 5
    nc = 5
    mg = RasterModelGrid((nr, nc), xy_spacing=10.0)

    mg.add_zeros("topographic__elevation", at="node")

    mg["node"]["topographic__elevation"] += (
        mg.node_y / 100000 + mg.node_x / 100000 + np.random.rand(len(mg.node_y)) / 10000
    )
    mg.set_closed_boundaries_at_grid_edges(
        bottom_is_closed=True,
        left_is_closed=True,
        right_is_closed=True,
        top_is_closed=True,
    )
    mg.set_watershed_boundary_condition_outlet_id(
        0, mg["node"]["topographic__elevation"], -9999.0
    )

    # Create a D8 flow handler
    FlowAccumulator(
        mg, flow_director="D8", depression_finder="DepressionFinderAndRouter"
    )

    # Instantiate the ErosionDeposition component...
    with pytest.raises(ValueError):
        ErosionDeposition(
            mg,
            K=0.01,
            F_f=2.0,
            v_s=0.001,
            m_sp=0.5,
            n_sp=1.0,
            sp_crit=0.0,
            solver="basic",
        )


def test_Ff_too_low_vals():
    """
    Test that instantiating ErosionDeposition with a F_f value < 0 throws a
    ValueError.
    """

    # set up a 5x5 grid with one open outlet node and low initial elevations.
    nr = 5
    nc = 5
    mg = RasterModelGrid((nr, nc), xy_spacing=10.0)

    mg.add_zeros("topographic__elevation", at="node")

    mg["node"]["topographic__elevation"] += (
        mg.node_y / 100000 + mg.node_x / 100000 + np.random.rand(len(mg.node_y)) / 10000
    )
    mg.set_closed_boundaries_at_grid_edges(
        bottom_is_closed=True,
        left_is_closed=True,
        right_is_closed=True,
        top_is_closed=True,
    )
    mg.set_watershed_boundary_condition_outlet_id(
        0, mg["node"]["topographic__elevation"], -9999.0
    )

    # Create a D8 flow handler
    FlowAccumulator(
        mg, flow_director="D8", depression_finder="DepressionFinderAndRouter"
    )

    # Instantiate the ErosionDeposition component...
    with pytest.raises(ValueError):
        ErosionDeposition(
            mg,
            K=0.01,
            F_f=-0.5,
            v_s=0.001,
            m_sp=0.5,
            n_sp=1.0,
            sp_crit=0.0,
            solver="basic",
        )


def test_q_as_field():
    """
    Test that passing in water discharge as a grid field results in self.q
    holding correct values
    """

    # set up a 5x5 grid with one open outlet node and low initial elevations.
    nr = 5
    nc = 5
    mg = RasterModelGrid((nr, nc), xy_spacing=10.0)

    mg.add_zeros("topographic__elevation", at="node")
    q = mg.add_zeros("user_imposed_discharge", at="node")
    q[:] += 1.0  # add 1.0 m3/yr of water

    mg["node"]["topographic__elevation"] += (
        mg.node_y / 100000 + mg.node_x / 100000 + np.random.rand(len(mg.node_y)) / 10000
    )
    mg.set_closed_boundaries_at_grid_edges(
        bottom_is_closed=True,
        left_is_closed=True,
        right_is_closed=True,
        top_is_closed=True,
    )
    mg.set_watershed_boundary_condition_outlet_id(
        0, mg["node"]["topographic__elevation"], -9999.0
    )

    # Create a D8 flow handler
    FlowAccumulator(
        mg, flow_director="D8", depression_finder="DepressionFinderAndRouter"
    )

    # Instantiate the ErosionDeposition component...
    ed = ErosionDeposition(
        mg,
        K=0.01,
        F_f=0.0,
        v_s=0.001,
        m_sp=0.5,
        n_sp=1.0,
        sp_crit=0.0,
        discharge_field="user_imposed_discharge",
        solver="basic",
    )

    # ensure that ed._q is everywhere equal to 1.0 m3/yr.
    testing.assert_array_equal(
        np.ones(mg.number_of_nodes),
        ed._q,
        err_msg="E/D discharge field test failed",
        verbose=True,
    )


def test_q_as_array():
    """
    Test that passing in water discharge as an array results in self.q
    holding correct values
    """

    # set up a 5x5 grid with one open outlet node and low initial elevations.
    nr = 5
    nc = 5
    mg = RasterModelGrid((nr, nc), xy_spacing=10.0)

    mg.add_zeros("topographic__elevation", at="node")
    q = np.zeros(mg.number_of_nodes)
    q[:] += 1.0  # add 1.0 m3/yr of water

    mg["node"]["topographic__elevation"] += (
        mg.node_y / 100000 + mg.node_x / 100000 + np.random.rand(len(mg.node_y)) / 10000
    )
    mg.set_closed_boundaries_at_grid_edges(
        bottom_is_closed=True,
        left_is_closed=True,
        right_is_closed=True,
        top_is_closed=True,
    )
    mg.set_watershed_boundary_condition_outlet_id(
        0, mg["node"]["topographic__elevation"], -9999.0
    )

    # Create a D8 flow handler
    FlowAccumulator(
        mg, flow_director="D8", depression_finder="DepressionFinderAndRouter"
    )

    # Instantiate the ErosionDeposition component...
    ed = ErosionDeposition(
        mg,
        K=0.01,
        F_f=0.0,
        v_s=0.001,
        m_sp=0.5,
        n_sp=1.0,
        sp_crit=0.0,
        discharge_field=q,
        solver="basic",
    )

    # ensure that ed._q is everywhere equal to 1.0 m3/yr.
    testing.assert_array_equal(
        np.ones(mg.number_of_nodes),
        ed._q,
        err_msg="E/D discharge array test failed",
        verbose=True,
    )


def test_sediment__outflux_already_created():
    """
    Test that an existing sediment flux grid field is not changed by
    instantiating ErosionDeposition.
    """

    # set up a 5x5 grid with one open outlet node and low initial elevations.
    nr = 5
    nc = 5
    mg = RasterModelGrid((nr, nc), xy_spacing=10.0)

    mg.add_zeros("topographic__elevation", at="node")
    qs = mg.add_zeros("sediment__outflux", at="node")
    qs[:] += 1.0  # add 1.0 m3/yr of flux

    mg["node"]["topographic__elevation"] += (
        mg.node_y / 100000 + mg.node_x / 100000 + np.random.rand(len(mg.node_y)) / 10000
    )
    mg.set_closed_boundaries_at_grid_edges(
        bottom_is_closed=True,
        left_is_closed=True,
        right_is_closed=True,
        top_is_closed=True,
    )
    mg.set_watershed_boundary_condition_outlet_id(
        0, mg["node"]["topographic__elevation"], -9999.0
    )

    # Create a D8 flow handler
    FlowAccumulator(
        mg, flow_director="D8", depression_finder="DepressionFinderAndRouter"
    )

    # Instantiate the ErosionDeposition component...
    ed = ErosionDeposition(
        mg,
        K=0.01,
        F_f=0.0,
        v_s=0.001,
        m_sp=0.5,
        n_sp=1.0,
        sp_crit=0.0,
        solver="basic",
    )

    # ensure that 'sediment__flux' field is everywhere equal to 1.0 m3/yr.
    testing.assert_array_equal(
        np.ones(mg.number_of_nodes),
        ed._qs,
        err_msg="E/D sediment flux field test failed",
        verbose=True,
    )
